"""
Drawing routines to draw graphs.

This module contains routines to draw graphs on:

    - Cairo surfaces (L{DefaultGraphDrawer})
    - Matplotlib axes (L{MatplotlibGraphDrawer})

It also contains routines to send an igraph graph directly to
(U{Cytoscape<https://cytoscape.org>}) using the
(U{CytoscapeRPC plugin<https://apps.cytoscape.org/apps/cytoscaperpc>}), see
L{CytoscapeGraphDrawer}. L{CytoscapeGraphDrawer} can also fetch the current
network from Cytoscape and convert it to igraph format.
"""

from warnings import warn

from igraph.drawing.baseclasses import AbstractGraphDrawer, AbstractXMLRPCDrawer

__all__ = ("CytoscapeGraphDrawer", "__plot__")


class CytoscapeGraphDrawer(AbstractXMLRPCDrawer, AbstractGraphDrawer):
    """Graph drawer that sends/receives graphs to/from Cytoscape using
    CytoscapeRPC.

    This graph drawer cooperates with U{Cytoscape<https://cytoscape.org>}
    using U{CytoscapeRPC<https://apps.cytoscape.org/apps/cytoscaperpc>}.
    You need to install the CytoscapeRPC plugin first and start the
    XML-RPC server on a given port (port 9000 by default) from the
    appropriate Plugins submenu in Cytoscape.

    Graph, vertex and edge attributes are transferred to Cytoscape whenever
    possible (i.e. when a suitable mapping exists between a Python type
    and a Cytoscape type). If there is no suitable Cytoscape type for a
    Python type, the drawer will use a string attribute on the Cytoscape
    side and invoke C{str()} on the Python attributes.

    If an attribute to be created on the Cytoscape side already exists with
    a different type, an underscore will be appended to the attribute name
    to resolve the type conflict.

    You can use the C{network_id} attribute of this class to figure out the
    network ID of the last graph drawn with this drawer.
    """

    def __init__(self, url="http://localhost:9000/Cytoscape"):
        """Constructs a Cytoscape graph drawer using the XML-RPC interface
        of Cytoscape at the given URL."""
        super().__init__(url, "Cytoscape")
        self.network_id = None

    def draw(self, graph, name="Network from igraph", create_view=True, *args, **kwds):
        """Sends the given graph to Cytoscape as a new network.

        @param name: the name of the network in Cytoscape.
        @param create_view: whether to create a view for the network
          in Cytoscape.The default is C{True}.
        @keyword node_ids: specifies the identifiers of the nodes to
          be used in Cytoscape. This must either be the name of a
          vertex attribute or a list specifying the identifiers, one
          for each node in the graph. The default is C{None}, which
          simply uses the vertex index for each vertex."""
        from xmlrpc.client import Fault

        # Positional arguments are not used
        if args:
            warn(
                "Positional arguments to plot functions are ignored "
                "and will be deprecated soon.",
                DeprecationWarning,
                stacklevel=1,
            )

        cy = self.service

        # Create the network
        if not create_view:
            try:
                network_id = cy.createNetwork(name, False)
            except Fault:
                warn(
                    "CytoscapeRPC too old, cannot create network without view."
                    " Consider upgrading CytoscapeRPC to use this feature.",
                    stacklevel=1,
                )
                network_id = cy.createNetwork(name)
        else:
            network_id = cy.createNetwork(name)
        self.network_id = network_id

        # Create the nodes
        if "node_ids" in kwds:
            node_ids = kwds["node_ids"]
            if isinstance(node_ids, str):
                node_ids = graph.vs[node_ids]
        else:
            node_ids = list(range(graph.vcount()))
        node_ids = [str(identifier) for identifier in node_ids]
        cy.createNodes(network_id, node_ids)

        # Create the edges
        edgelists = [[], []]
        for v1, v2 in graph.get_edgelist():
            edgelists[0].append(node_ids[v1])
            edgelists[1].append(node_ids[v2])
        edge_ids = cy.createEdges(
            network_id,
            edgelists[0],
            edgelists[1],
            ["unknown"] * graph.ecount(),
            [graph.is_directed()] * graph.ecount(),
            False,
        )

        if "layout" in kwds:
            # Calculate/get the layout of the graph
            layout = self.ensure_layout(kwds["layout"], graph)
            size = 100 * graph.vcount() ** 0.5
            layout.fit_into((size, size), keep_aspect_ratio=True)
            layout.translate(-size / 2.0, -size / 2.0)
            cy.setNodesPositions(network_id, node_ids, *list(zip(*list(layout))))
        else:
            # Ask Cytoscape to perform the default layout so the user can
            # at least see something in Cytoscape while the attributes are
            # being transferred
            cy.performDefaultLayout(network_id)

        # Send the network attributes
        attr_names = set(cy.getNetworkAttributeNames())
        for attr in graph.attributes():
            cy_type, value = self.infer_cytoscape_type([graph[attr]])
            value = value[0]
            if value is None:
                continue

            # Resolve type conflicts (if any)
            try:
                while (
                    attr in attr_names and cy.getNetworkAttributeType(attr) != cy_type
                ):
                    attr += "_"
            except Fault:
                # getNetworkAttributeType is not available in some older versions
                # so we simply pass here
                pass
            cy.addNetworkAttributes(attr, cy_type, {network_id: value})

        # Send the node attributes
        attr_names = set(cy.getNodeAttributeNames())
        for attr in graph.vertex_attributes():
            cy_type, values = self.infer_cytoscape_type(graph.vs[attr])
            values = dict(pair for pair in zip(node_ids, values) if pair[1] is not None)
            # Resolve type conflicts (if any)
            while attr in attr_names and cy.getNodeAttributeType(attr) != cy_type:
                attr += "_"
            # Send the attribute values
            cy.addNodeAttributes(attr, cy_type, values, True)

        # Send the edge attributes
        attr_names = set(cy.getEdgeAttributeNames())
        for attr in graph.edge_attributes():
            cy_type, values = self.infer_cytoscape_type(graph.es[attr])
            values = dict(pair for pair in zip(edge_ids, values) if pair[1] is not None)
            # Resolve type conflicts (if any)
            while attr in attr_names and cy.getEdgeAttributeType(attr) != cy_type:
                attr += "_"
            # Send the attribute values
            cy.addEdgeAttributes(attr, cy_type, values)

    def fetch(self, name=None, directed=False, keep_canonical_names=False):
        """Fetches the network with the given name from Cytoscape.

        When fetching networks from Cytoscape, the C{canonicalName} attributes
        of vertices and edges are not converted by default. Use the
        C{keep_canonical_names} parameter to retrieve these attributes as well.

        @param name: the name of the network in Cytoscape.
        @param directed: whether the network is directed.
        @param keep_canonical_names: whether to keep the C{canonicalName}
            vertex/edge attributes that are added automatically by Cytoscape
        @return: an appropriately constructed igraph L{Graph}."""
        from igraph import Graph

        cy = self.service

        # Check the version number. Anything older than 1.3 is bad.
        version = cy.version()
        if " " in version:
            version = version.split(" ")[0]
        version = tuple(map(int, version.split(".")[:2]))
        if version < (1, 3):
            raise NotImplementedError(
                "CytoscapeGraphDrawer requires Cytoscape-RPC 1.3 or newer"
            )

        # Find out the ID of the network we are interested in
        if name is None:
            network_id = cy.getNetworkID()
        else:
            network_id = [k for k, v in cy.getNetworkList().items() if v == name]
            if not network_id:
                raise ValueError("no such network: %r" % name)
            elif len(network_id) > 1:
                raise ValueError("more than one network exists with name: %r" % name)
            network_id = network_id[0]

        # Fetch the list of all the nodes and edges
        vertices = cy.getNodes(network_id)
        edges = cy.getEdges(network_id)
        n, m = len(vertices), len(edges)

        # Fetch the graph attributes
        graph_attrs = cy.getNetworkAttributes(network_id)

        # Fetch the vertex attributes
        vertex_attr_names = cy.getNodeAttributeNames()
        vertex_attrs = {}
        for attr_name in vertex_attr_names:
            if attr_name == "canonicalName" and not keep_canonical_names:
                continue
            has_attr = cy.nodesHaveAttribute(attr_name, vertices)
            filtered = [idx for idx, ok in enumerate(has_attr) if ok]
            values = cy.getNodesAttributes(
                attr_name, [name for name, ok in zip(vertices, has_attr) if ok]
            )
            attrs = [None] * n
            for idx, value in zip(filtered, values):
                attrs[idx] = value
            vertex_attrs[attr_name] = attrs

        # Fetch the edge attributes
        edge_attr_names = cy.getEdgeAttributeNames()
        edge_attrs = {}
        for attr_name in edge_attr_names:
            if attr_name == "canonicalName" and not keep_canonical_names:
                continue
            has_attr = cy.edgesHaveAttribute(attr_name, edges)
            filtered = [idx for idx, ok in enumerate(has_attr) if ok]
            values = cy.getEdgesAttributes(
                attr_name, [name for name, ok in zip(edges, has_attr) if ok]
            )
            attrs = [None] * m
            for idx, value in zip(filtered, values):
                attrs[idx] = value
            edge_attrs[attr_name] = attrs

        # Create a vertex name index
        vertex_name_index = {v: k for k, v in enumerate(vertices)}
        del vertices

        # Remap the edges list to numeric IDs
        edge_list = []
        for edge in edges:
            parts = edge.split()
            edge_list.append((vertex_name_index[parts[0]], vertex_name_index[parts[2]]))
        del edges

        return Graph(
            n,
            edge_list,
            directed=directed,
            graph_attrs=graph_attrs,
            vertex_attrs=vertex_attrs,
            edge_attrs=edge_attrs,
        )

    @staticmethod
    def infer_cytoscape_type(values):
        """Returns a Cytoscape type that can be used to represent all the
        values in C{values} and an appropriately converted copy of C{values} that
        is suitable for an XML-RPC call.  Note that the string type in
        Cytoscape is used as a catch-all type; if no other type fits, attribute
        values will be converted to string and then posted to Cytoscape.

        C{None} entries are allowed in C{values}, they will be ignored on the
        Cytoscape side.
        """
        types = [type(value) for value in values if value is not None]
        if all(t == bool for t in types):
            return "BOOLEAN", values
        if all(issubclass(t, (int, int)) for t in types):
            return "INTEGER", values
        if all(issubclass(t, float) for t in types):
            return "FLOATING", values
        return "STRING", [
            str(value) if not isinstance(value, str) else value for value in values
        ]


#####################################################################


class GephiGraphStreamingDrawer(AbstractGraphDrawer):
    """Graph drawer that sends a graph to a file-like object (e.g., socket, URL
    connection, file) using the Gephi graph streaming format.

    The Gephi graph streaming format is a simple JSON-based format that can be used
    to post mutations to a graph (i.e. node and edge additions, removals and updates)
    to a remote component. For instance, one can open up Gephi
    (U{http://www.gephi.org}), install the Gephi graph streaming plugin and then
    send a graph from igraph straight into the Gephi window by using
    C{GephiGraphStreamingDrawer} with the appropriate URL where Gephi is
    listening.

    The C{connection} property exposes the L{GephiConnection} that the drawer
    uses. The drawer also has a property called C{streamer} which exposes the underlying
    L{GephiGraphStreamer} that is responsible for generating the JSON objects,
    encoding them and writing them to a file-like object. If you want to customize
    the encoding process, this is the object where you can tweak things to your taste.
    """

    def __init__(self, conn=None, *args, **kwds):
        """Constructs a Gephi graph streaming drawer that will post graphs to the
        given Gephi connection. If C{conn} is C{None}, the remaining arguments of
        the constructor are forwarded intact to the constructor of
        L{GephiConnection} in order to create a connection. This means that any of
        the following are valid:

          - C{GephiGraphStreamingDrawer()} will construct a drawer that connects to
            workspace 0 of the local Gephi instance on port 8080.

          - C{GephiGraphStreamingDrawer(workspace=2)} will connect to workspace 2
            of the local Gephi instance on port 8080.

          - C{GephiGraphStreamingDrawer(port=1234)} will connect to workspace 0
            of the local Gephi instance on port 1234.

          - C{GephiGraphStreamingDrawer(host="remote", port=1234, workspace=7)}
            will connect to workspace 7 of the Gephi instance on host C{remote},
            port 1234.

          - C{GephiGraphStreamingDrawer(url="http://remote:1234/workspace7)} is
            the same as above, but with an explicit URL.
        """
        super().__init__()

        from igraph.remote.gephi import GephiGraphStreamer, GephiConnection

        self.connection = conn or GephiConnection(*args, **kwds)
        self.streamer = GephiGraphStreamer()

    def draw(self, graph, *args, **kwds):
        """Draws (i.e. sends) the given graph to the destination of the drawer using
        the Gephi graph streaming API.

        The following keyword arguments are allowed:

            - C{encoder} lets one specify an instance of C{json.JSONEncoder} that
              will be used to encode the JSON objects.
        """
        # Positional arguments are not used
        if args:
            warn(
                "Positional arguments to plot functions are ignored "
                "and will be deprecated soon.",
                DeprecationWarning,
                stacklevel=1,
            )

        self.streamer.post(graph, self.connection, encoder=kwds.get("encoder"))


def __plot__(self, backend, context, *args, **kwds):
    """Plots the graph to the given Cairo context or matplotlib Axes.

    The visual style of vertices and edges can be modified at three
    places in the following order of precedence (lower indices override
    higher indices):

      1. Keyword arguments of this function (or of L{plot()} which is
         passed intact to C{Graph.__plot__()}.

      2. Vertex or edge attributes, specified later in the list of
         keyword arguments.

      3. igraph-wide plotting defaults (see
         L{igraph.config.Configuration})

      4. Built-in defaults.

    E.g., if the C{vertex_size} keyword attribute is not present,
    but there exists a vertex attribute named C{size}, the sizes of
    the vertices will be specified by that attribute.

    Besides the usual self-explanatory plotting parameters (C{context},
    C{bbox}, C{palette}), it accepts the following keyword arguments:

      - C{autocurve}: whether to use curves instead of straight lines for
        multiple edges on the graph plot. This argument may be C{True}
        or C{False}; when omitted, C{True} is assumed for graphs with
        less than 10.000 edges and C{False} otherwise.

      - C{drawer_factory}: a subclass of L{AbstractCairoGraphDrawer}
        which will be used to draw the graph. You may also provide
        a function here which takes two arguments: the Cairo context
        to draw on and a bounding box (an instance of L{BoundingBox}).
        If this keyword argument is missing, igraph will use the
        default graph drawer which should be suitable for most purposes.
        It is safe to omit this keyword argument unless you need to use
        a specific graph drawer.

      - C{keep_aspect_ratio}: whether to keep the aspect ratio of the layout
        that igraph calculates to place the nodes. C{True} means that the
        layout will be scaled proportionally to fit into the bounding box
        where the graph is to be drawn but the aspect ratio will be kept
        the same (potentially leaving empty space next to, below or above
        the graph). C{False} means that the layout will be scaled independently
        along the X and Y axis in order to fill the entire bounding box.
        The default is C{False}.

      - C{layout}: the layout to be used. If not an instance of
        L{Layout}, it will be passed to L{layout} to calculate
        the layout. Note that if you want a deterministic layout that
        does not change with every plot, you must either use a
        deterministic layout function (like L{GraphBase.layout_circle}) or
        calculate the layout in advance and pass a L{Layout} object here.

      - C{margin}: the top, right, bottom, left margins as a 4-tuple.
        If it has less than 4 elements or is a single float, the elements
        will be re-used until the length is at least 4.

      - C{mark_groups}: whether to highlight some of the vertex groups by
        colored polygons. This argument can be one of the following:

          - C{False}: no groups will be highlighted

          - C{True}: only valid if the object plotted is a
            L{VertexClustering} or L{VertexCover}. The vertex groups in the
            clutering or cover will be highlighted such that the i-th
            group will be colored by the i-th color from the current
            palette. If used when plotting a graph, it will throw an error.

          - A dict mapping tuples of vertex indices to color names.
            The given vertex groups will be highlighted by the given
            colors.

          - A list containing pairs or an iterable yielding pairs, where
            the first element of each pair is a list of vertex indices and
            the second element is a color.

          - A L{VertexClustering} or L{VertexCover} instance. The vertex
            groups in the clustering or cover will be highlighted such that
            the i-th group will be colored by the i-th color from the
            current palette.

        In place of lists of vertex indices, you may also use L{VertexSeq}
        instances.

        In place of color names, you may also use color indices into the
        current palette. C{None} as a color name will mean that the
        corresponding group is ignored.

      - C{vertex_size}: size of the vertices. The corresponding vertex
        attribute is called C{size}. The default is 10. Vertex sizes
        are measured in the unit of the Cairo context on which igraph
        is drawing.

      - C{vertex_color}: color of the vertices. The corresponding vertex
        attribute is C{color}, the default is red.  Colors can be
        specified either by common X11 color names (see the source
        code of L{igraph.drawing.colors} for a list of known colors), by
        3-tuples of floats (ranging between 0 and 255 for the R, G and
        B components), by CSS-style string specifications (C{#rrggbb})
        or by integer color indices of the specified palette.

      - C{vertex_frame_color}: color of the frame (i.e. stroke) of the
        vertices. The corresponding vertex attribute is C{frame_color},
        the default is black. See C{vertex_color} for the possible ways
        of specifying a color.

      - C{vertex_frame_width}: the width of the frame (i.e. stroke) of the
        vertices. The corresponding vertex attribute is C{frame_width}.
        The default is 1. Vertex frame widths are measured in the unit of the
        Cairo context on which igraph is drawing.

      - C{vertex_shape}: shape of the vertices. Alternatively it can
        be specified by the C{shape} vertex attribute. Possibilities
        are: C{square}, {circle}, {triangle}, {triangle-down} or
        C{hidden}. See the source code of L{igraph.drawing} for a
        list of alternative shape names that are also accepted and
        mapped to these.

      - C{vertex_label}: labels drawn next to the vertices.
        The corresponding vertex attribute is C{label}.

      - C{vertex_label_dist}: distance of the midpoint of the vertex
        label from the center of the corresponding vertex.
        The corresponding vertex attribute is C{label_dist}.

      - C{vertex_label_color}: color of the label. Corresponding
        vertex attribute: C{label_color}. See C{vertex_color} for
        color specification syntax.

      - C{vertex_label_size}: font size of the label, specified
        in the unit of the Cairo context on which we are drawing.
        Corresponding vertex attribute: C{label_size}.

      - C{vertex_label_angle}: the direction of the line connecting
        the midpoint of the vertex with the midpoint of the label.
        This can be used to position the labels relative to the
        vertices themselves in conjunction with C{vertex_label_dist}.
        Corresponding vertex attribute: C{label_angle}. The
        default is C{-math.pi/2}.

      - C{vertex_order}: drawing order of the vertices. This must be
        a list or tuple containing vertex indices; vertices are then
        drawn according to this order.

      - C{vertex_order_by}: an alternative way to specify the drawing
        order of the vertices; this attribute is interpreted as the name
        of a vertex attribute, and vertices are drawn such that those
        with a smaller attribute value are drawn first. You may also
        reverse the order by passing a tuple here; the first element of
        the tuple should be the name of the attribute, the second element
        specifies whether the order is reversed (C{True}, C{False},
        C{"asc"} and C{"desc"} are accepted values).

      - C{edge_color}: color of the edges. The corresponding edge
        attribute is C{color}, the default is red. See C{vertex_color}
        for color specification syntax.

      - C{edge_curved}: whether the edges should be curved. Positive
        numbers correspond to edges curved in a counter-clockwise
        direction, negative numbers correspond to edges curved in a
        clockwise direction. Zero represents straight edges. C{True}
        is interpreted as 0.5, C{False} is interpreted as 0. The
        default is 0 which makes all the edges straight.

      - C{edge_width}: width of the edges in the default unit of the
        Cairo context on which we are drawing. The corresponding
        edge attribute is C{width}, the default is 1.

      - C{edge_arrow_size}: arrow size of the edges. The
        corresponding edge attribute is C{arrow_size}, the default
        is 1.

      - C{edge_arrow_width}: width of the arrowhead on the edge. The
        corresponding edge attribute is C{arrow_width}, the default
        is 1.

      - C{edge_order}: drawing order of the edges. This must be
        a list or tuple containing edge indices; edges are then
        drawn according to this order.

      - C{edge_order_by}: an alternative way to specify the drawing
        order of the edges; this attribute is interpreted as the name
        of an edge attribute, and edges are drawn such that those
        with a smaller attribute value are drawn first. You may also
        reverse the order by passing a tuple here; the first element of
        the tuple should be the name of the attribute, the second element
        specifies whether the order is reversed (C{True}, C{False},
        C{"asc"} and C{"desc"} are accepted values).
    """
    from igraph.drawing import DrawerDirectory

    drawer = kwds.pop(
        "drawer_factory",
        DrawerDirectory.resolve(self, backend)(context),
    )
    return drawer.draw(self, *args, **kwds)



"""
.. _tutorials-quickstart:

===========
Quick Start
===========
For the eager folks out there, this intro will give you a quick overview of the following operations:

- Construct a graph
- Set attributes of nodes and edges
- Plot a graph using matplotlib
- Save the plot as an image
- Export and import a graph as a ``.gml`` file

To find out more features that igraph has to offer, check out the :ref:`gallery`!

"""

import igraph as ig
import matplotlib.pyplot as plt

# Construct a graph with 5 vertices
n_vertices = 5
edges = [(0, 1), (0, 2), (0, 3), (0, 4), (1, 2), (1, 3), (1, 4), (3, 4)]
g = ig.Graph(n_vertices, edges)

# Set attributes for the graph, nodes, and edges
g["title"] = "Small Social Network"
g.vs["name"] = [
    "Daniel Morillas",
    "Kathy Archer",
    "Kyle Ding",
    "Joshua Walton",
    "Jana Hoyer",
]
g.vs["gender"] = ["M", "F", "F", "M", "F"]
g.es["married"] = [False, False, False, False, False, False, False, True]

# Set individual attributes
g.vs[1]["name"] = "Kathy Morillas"
g.es[0]["married"] = True

# Plot in matplotlib
# Note that attributes can be set globally (e.g. vertex_size), or set individually using arrays (e.g. vertex_color)
fig, ax = plt.subplots(figsize=(5, 5))
ig.plot(
    g,
    target=ax,
    layout="circle",  # print nodes in a circular layout
    vertex_size=30,
    vertex_color=[
        "steelblue" if gender == "M" else "salmon" for gender in g.vs["gender"]
    ],
    vertex_frame_width=4.0,
    vertex_frame_color="white",
    vertex_label=g.vs["name"],
    vertex_label_size=7.0,
    edge_width=[2 if married else 1 for married in g.es["married"]],
    edge_color=["#7142cf" if married else "#AAA" for married in g.es["married"]],
)

plt.show()

# Save the graph as an image file
fig.savefig("social_network.png")
fig.savefig("social_network.jpg")
fig.savefig("social_network.pdf")

# Export and import a graph as a GML file.
g.save("social_network.gml")
g = ig.load("social_network.gml")



"""
Drawing routines to draw the matrices.

This module provides implementations of matrix drawers.
"""

from igraph.drawing.baseclasses import AbstractDrawer
from igraph.drawing.utils import str_to_orientation

from .utils import find_matplotlib

__all__ = ("MatplotlibDendrogramDrawer",)

mpl, _ = find_matplotlib()


class MatplotlibDendrogramDrawer(AbstractDrawer):
    """Matplotlib drawer object for dendrograms."""

    def __init__(self, ax):
        """Constructs the drawer and associates it to the given Axes.

        @param ax: the Axes on which we will draw
        """
        super().__init__()
        self.context = ax

    def _plot_item(self, dendro, ax, orientation, idx, x, y):
        """Plots a dendrogram item to the given Cairo context

        @param dendro: the dendrogram object
        @param orientation: whether the dendrogram is horizontally oriented
        @param idx: the index of the item
        @param x: the X position of the item
        @param y: the Y position of the item
        """
        if dendro._names is None or dendro._names[idx] is None:
            return

        if orientation == "lr":
            ha, va, rotation = "right", "center", 0
        elif orientation == "rl":
            ha, va, rotation = "left", "center", 0
        elif orientation == "tb":
            ha, va, rotation = "center", "bottom", 90
        else:
            ha, va, rotation = "center", "top", 90

        # TODO: offset a little? But remember zoom in callbacks

        ax.text(
            x,
            y,
            dendro._names[idx],
            ha=ha,
            va=va,
            rotation=rotation,
        )

    def draw(self, dendro, orientation="lr", **kwds):
        """Draws the given Dendrogram in a matplotlib Axes.

        Other keyword arguments are passed to mpl.patches.Polygon.

        @param dendro: the igraph.Dendrogram to plot.
        @param orientation: the direction of the plot. Accepted values are "lr"
          (root on the right), "rl" (root on the left), "tb" (root at the bottom),
          and "bt" (root at the top). A few aliases are available (see
          L{utils.str_to_orientation}).
        """
        from igraph.layout import Layout

        ax = self.context

        # Pop unneeded arguments from kwds that are passed down to us by
        # default but cannot be interpreted by Matplotlib
        kwds.pop("palette", None)

        # Styling defaults
        kwds["edgecolor"] = kwds.pop("color", "black")
        if ("lw" not in kwds) and ("linewidth" not in kwds):
            kwds["linewidth"] = 1

        if dendro._names is None:
            dendro._names = [str(x) for x in range(dendro._nitems)]

        orientation = str_to_orientation(orientation, reversed_vertical=True)
        horiz = orientation in ("lr", "rl")

        # Calculate node coordinates
        layout = Layout([(0, 0)] * dendro._nitems, dim=2)
        inorder = dendro._traverse_inorder()
        if not horiz:
            x, y = 0, 0
            # Leaves
            for element in inorder:
                layout[element] = (x, 0)
                x += 1

            # Internal nodes
            for id1, id2 in dendro._merges:
                x = (layout[id1][0] + layout[id2][0]) / 2.0
                # TODO: this is a little restrictive, but alright
                # for such a simple layout. More complex tree layouts
                # should be in a separate Layout anyway
                y += 1
                layout.append((x, y))

            # Mirror or rotate the layout if necessary
            if orientation == "bt":
                layout.mirror(1)
        else:
            x, y = 0, 0
            for element in inorder:
                layout[element] = (0, y)
                y += 1

            for id1, id2 in dendro._merges:
                y = (layout[id1][1] + layout[id2][1]) / 2.0
                # TODO: this is a little restrictive, but alright
                # for such a simple layout. More complex tree layouts
                # should be in a separate Layout anyway
                x += 1
                layout.append((x, y))

            # Mirror or rotate the layout if necessary
            if orientation == "rl":
                layout.mirror(0)

        # Draw leaf names
        #
        # for idx in range(dendro._nitems):
        #    x, y = layout[idx]
        #    self._plot_item(dendro, ax, orientation, idx, x, y)
        ticks, ticklabels = [], []
        for idx in range(dendro._nitems):
            x, y = layout[idx]
            if not horiz:
                ticks.append(x)
            else:
                ticks.append(y)
            ticklabels.append(dendro._names[idx])

        if not horiz:
            ax.set_xticks(ticks)
            ax.set_xticklabels(ticklabels)
            ax.set_yticks([])
        else:
            ax.set_yticks(ticks)
            ax.set_yticklabels(ticklabels)
            ax.set_xticks([])

        # Draw dendrogram lines
        # Each path is a U-shaped fork
        if not horiz:
            for idx, (id1, id2) in enumerate(dendro._merges):
                x0, y0 = layout[id1]
                x1, y1 = layout[id2]
                x2, y2 = layout[idx + dendro._nitems]
                poly = mpl.patches.Polygon(
                    [[x0, y0], [x0, y2], [x1, y2], [x1, y1]],
                    closed=False,
                    facecolor="none",
                    **kwds,
                )
                ax.add_patch(poly)
        else:
            for idx, (id1, id2) in enumerate(dendro._merges):
                x0, y0 = layout[id1]
                x1, y1 = layout[id2]
                x2, y2 = layout[idx + dendro._nitems]
                poly = mpl.patches.Polygon(
                    [[x0, y0], [x2, y0], [x2, y1], [x1, y1]],
                    closed=False,
                    facecolor="none",
                    **kwds,
                )
                ax.add_patch(poly)

        ax.autoscale_view()



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""Classes representing cuts and flows on graphs."""

from igraph._igraph import (
    GraphBase,
)
from igraph.clustering import VertexClustering


class Cut(VertexClustering):
    """A cut of a given graph.

    This is a simple class used to represent cuts returned by
    L{Graph.mincut()}, L{Graph.all_st_cuts()} and other functions
    that calculate cuts.

    A cut is a special vertex clustering with only two clusters.
    Besides the usual L{VertexClustering} methods, it also has the
    following attributes:

      - C{value} - the value (capacity) of the cut. It is equal to
        the number of edges if there are no capacities on the
        edges.

      - C{partition} - vertex IDs in the parts created
        after removing edges in the cut

      - C{cut} - edge IDs in the cut

      - C{es} - an edge selector restricted to the edges
        in the cut.

    You can use indexing on this object to obtain lists of vertex IDs
    for both sides of the partition.

    This class is usually not instantiated directly, everything
    is taken care of by the functions that return cuts.

    Examples:

      >>> from igraph import Graph
      >>> g = Graph.Ring(20)
      >>> mc = g.mincut()
      >>> print(mc.value)
      2.0
      >>> print(min(len(x) for x in mc))
      1
      >>> mc.es["color"] = "red"
    """

    def __init__(self, graph, value=None, cut=None, partition=None, partition2=None):
        """Initializes the cut.

        This should not be called directly, everything is taken care of by
        the functions that return cuts.
        """
        # Input validation
        if partition is None or cut is None:
            raise ValueError("partition and cut must be given")

        # Set up a membership vector, initialize parent class
        membership = [1] * graph.vcount()
        for vid in partition:
            membership[vid] = 0
        super().__init__(graph, membership)

        if value is None:
            # Value of the cut not given, count the number of edges
            value = len(cut)
        self._value = float(value)

        self._partition = sorted(partition)
        self._cut = cut

    def __repr__(self):
        return "%s(%r, %r, %r, %r)" % (
            self.__class__.__name__,
            self._graph,
            self._value,
            self._cut,
            self._partition,
        )

    def __str__(self):
        return "Graph cut (%d edges, %d vs %d vertices, value=%.4f)" % (
            len(self._cut),
            len(self._partition),
            self._graph.vcount() - len(self._partition),
            self._value,
        )

    @property
    def es(self):
        """Returns an edge selector restricted to the cut"""
        return self._graph.es.select(self.cut)

    @property
    def partition(self):
        """Returns the vertex IDs partitioned according to the cut"""
        return list(self)

    @property
    def cut(self):
        """Returns the edge IDs in the cut"""
        return self._cut

    @property
    def value(self):
        """Returns the sum of edge capacities in the cut"""
        return self._value


class Flow(Cut):
    """A flow of a given graph.

    This is a simple class used to represent flows returned by
    L{Graph.maxflow}. It has the following attributes:

      - C{graph} - the graph on which this flow is defined

      - C{value} - the value (capacity) of the flow

      - C{flow} - the flow values on each edge. For directed graphs,
        this is simply a list where element M{i} corresponds to the
        flow on edge M{i}. For undirected graphs, the direction of
        the flow is not constrained (since the edges are undirected),
        hence positive flow always means a flow from the smaller vertex
        ID to the larger, while negative flow means a flow from the
        larger vertex ID to the smaller.

      - C{cut} - edge IDs in the minimal cut corresponding to
        the flow.

      - C{partition} - vertex IDs in the parts created
        after removing edges in the cut

      - C{es} - an edge selector restricted to the edges
        in the cut.

    This class is usually not instantiated directly, everything
    is taken care of by L{Graph.maxflow}.

    Examples:

      >>> from igraph import Graph
      >>> g = Graph.Ring(20)
      >>> mf = g.maxflow(0, 10)
      >>> print(mf.value)
      2.0
      >>> mf.es["color"] = "red"
    """

    def __init__(self, graph, value, flow, cut, partition):
        """Initializes the flow.

        This should not be called directly, everything is
        taken care of by L{Graph.maxflow}.
        """
        super().__init__(graph, value, cut, partition)
        self._flow = flow

    def __repr__(self):
        return "%s(%r, %r, %r, %r, %r)" % (
            self.__class__.__name__,
            self._graph,
            self._value,
            self._flow,
            self._cut,
            self._partition,
        )

    def __str__(self):
        return "Graph flow (%d edges, %d vs %d vertices, value=%.4f)" % (
            len(self._cut),
            len(self._partition),
            self._graph.vcount() - len(self._partition),
            self._value,
        )

    @property
    def flow(self):
        """Returns the flow values for each edge.

        For directed graphs, this is simply a list where element M{i}
        corresponds to the flow on edge M{i}. For undirected graphs, the
        direction of the flow is not constrained (since the edges are
        undirected), hence positive flow always means a flow from the smaller
        vertex ID to the larger, while negative flow means a flow from the
        larger vertex ID to the smaller.
        """
        return self._flow


def _all_st_cuts(graph, source, target):
    """\
    Returns all the cuts between the source and target vertices in a
    directed graph.

    This function lists all edge-cuts between a source and a target vertex.
    Every cut is listed exactly once.

    B{Reference}: JS Provan and DR Shier: A paradigm for listing (s,t)-cuts in
    graphs. I{Algorithmica} 15, 351-372, 1996.

    @param source: the source vertex ID
    @param target: the target vertex ID
    @return: a list of L{Cut} objects.
    """
    return [
        Cut(graph, cut=cut, partition=part)
        for cut, part in zip(*GraphBase.all_st_cuts(graph, source, target))
    ]


def _all_st_mincuts(graph, source, target, capacity=None):
    """\
    Returns all the mincuts between the source and target vertices in a
    directed graph.

    This function lists all minimum edge-cuts between a source and a target
    vertex.

    B{Reference}: JS Provan and DR Shier: A paradigm for listing (s,t)-cuts in
    graphs. I{Algorithmica} 15, 351-372, 1996.

    @param source: the source vertex ID
    @param target: the target vertex ID
    @param capacity: the edge capacities (weights). If C{None}, all
      edges have equal weight. May also be an attribute name.
    @return: a list of L{Cut} objects.
    """
    value, cuts, parts = GraphBase.all_st_mincuts(graph, source, target, capacity)
    return [
        Cut(graph, value, cut=cut, partition=part) for cut, part in zip(cuts, parts)
    ]


def _gomory_hu_tree(graph, capacity=None, flow="flow"):
    """Calculates the Gomory-Hu tree of an undirected graph with optional
    edge capacities.

    The Gomory-Hu tree is a concise representation of the value of all the
    maximum flows (or minimum cuts) in a graph. The vertices of the tree
    correspond exactly to the vertices of the original graph in the same order.
    Edges of the Gomory-Hu tree are annotated by flow values.  The value of
    the maximum flow (or minimum cut) between an arbitrary (u,v) vertex
    pair in the original graph is then given by the minimum flow value (i.e.
    edge annotation) along the shortest path between u and v in the
    Gomory-Hu tree.

    @param capacity: the edge capacities (weights). If C{None}, all
      edges have equal weight. May also be an attribute name.
    @param flow: the name of the edge attribute in the returned graph
      in which the flow values will be stored.
    @return: the Gomory-Hu tree as a L{Graph} object.
    """
    graph, flow_values = GraphBase.gomory_hu_tree(graph, capacity)
    graph.es[flow] = flow_values
    return graph


def _maxflow(graph, source, target, capacity=None):
    """Returns a maximum flow between the given source and target vertices
    in a graph.

    A maximum flow from I{source} to I{target} is an assignment of
    non-negative real numbers to the edges of the graph, satisfying
    two properties:

        1. For each edge, the flow (i.e. the assigned number) is not
           more than the capacity of the edge (see the I{capacity}
           argument)

        2. For every vertex except the source and the target, the
           incoming flow is the same as the outgoing flow.

    The value of the flow is the incoming flow of the target or the
    outgoing flow of the source (which are equal). The maximum flow
    is the maximum possible such value.

    @param capacity: the edge capacities (weights). If C{None}, all
      edges have equal weight. May also be an attribute name.
    @return: a L{Flow} object describing the maximum flow
    """
    return Flow(graph, *GraphBase.maxflow(graph, source, target, capacity))


def _mincut(graph, source=None, target=None, capacity=None):
    """Calculates the minimum cut between the given source and target vertices
    or within the whole graph.

    The minimum cut is the minimum set of edges that needs to be removed to
    separate the source and the target (if they are given) or to disconnect the
    graph (if neither the source nor the target are given). The minimum is
    calculated using the weights (capacities) of the edges, so the cut with
    the minimum total capacity is calculated.

    For undirected graphs and no source and target, the method uses the
    Stoer-Wagner algorithm. For a given source and target, the method uses the
    push-relabel algorithm; see the references below.

    @param source: the source vertex ID. If C{None}, the target must also be
      C{None} and the calculation will be done for the entire graph (i.e.
      all possible vertex pairs).
    @param target: the target vertex ID. If C{None}, the source must also be
      C{None} and the calculation will be done for the entire graph (i.e.
      all possible vertex pairs).
    @param capacity: the edge capacities (weights). If C{None}, all
      edges have equal weight. May also be an attribute name.
    @return: a L{Cut} object describing the minimum cut
    """
    return Cut(graph, *GraphBase.mincut(graph, source, target, capacity))


def _st_mincut(graph, source, target, capacity=None):
    """Calculates the minimum cut between the source and target vertices in a
    graph.

    @param source: the source vertex ID
    @param target: the target vertex ID
    @param capacity: the capacity of the edges. It must be a list or a valid
      attribute name or C{None}. In the latter case, every edge will have the
      same capacity.
    @return: the value of the minimum cut, the IDs of vertices in the
      first and second partition, and the IDs of edges in the cut,
      packed in a 4-tuple
    """
    return Cut(graph, *GraphBase.st_mincut(graph, source, target, capacity))



from igraph._igraph import GraphBase

from .utils import deprecated

__all__ = ("_rewire", )


def _rewire(graph, n=None, allowed_edge_types="simple", *, mode=None):
    """Randomly rewires the graph while preserving the degree distribution.

    The rewiring is done \"in-place\", so the original graph will be modified.
    If you want to preserve the original graph, use the L{copy} method before
    rewiring.

    @param n: the number of rewiring trials. The default is 10 times the number
        of edges.
    @param allowed_edge_types: the rewiring algorithm to use. It can either be
        C{"simple"} or C{"loops"}; the former does not create or destroy
        loop edges while the latter does.
    """
    if mode is not None:
        deprecated("The 'mode' keyword argument is deprecated, use 'allowed_edge_types' instead")
        allowed_edge_types = mode

    return GraphBase._rewire(graph, n, allowed_edge_types)



"""
.. _tutorials-topological-sort:

===================
Topological sorting
===================

This example demonstrates how to get a topological sorting on a directed acyclic graph (DAG). A topological sorting of a directed graph is a linear ordering based on the precedence implied by the directed edges. It exists iff the graph doesn't have any cycle. In ``igraph``, we can use :meth:`igraph.GraphBase.topological_sorting` to get a topological ordering of the vertices.
"""

import igraph as ig
import matplotlib.pyplot as plt


# %%
# First off, we generate a directed acyclic graph (DAG):
g = ig.Graph(
    edges=[(0, 1), (0, 2), (1, 3), (2, 4), (4, 3), (3, 5), (4, 5)],
    directed=True,
)

# %%
# We can verify immediately that this is actually a DAG:
assert g.is_dag

# %%
# A topological sorting can be computed quite easily by calling
# :meth:`igraph.GraphBase.topological_sorting`, which returns a list of vertex IDs.
# If the given graph is not DAG, the error will occur.
results = g.topological_sorting(mode="out")
print("Topological sort of g (out):", *results)

# %%
# In fact, there are two modes of :meth:`igraph.GraphBase.topological_sorting`,
# ``'out'`` ``'in'``. ``'out'`` is the default and starts from a node with
# indegree equal to 0. Vice versa, ``'in'`` starts from a node with outdegree
# equal to 0. To call the other mode, we can simply use:
results = g.topological_sorting(mode="in")
print("Topological sort of g (in):", *results)

# %%
# We can use :meth:`igraph.Vertex.indegree` to find the indegree of the node.
for i in range(g.vcount()):
    print("degree of {}: {}".format(i, g.vs[i].indegree()))

# %
# Finally, we can plot the graph to make the situation a little clearer.
# Just to change things up a bit, we use the matplotlib visualization mode
# inspired by `xkcd <https://xkcd.com/>_:
with plt.xkcd():
    fig, ax = plt.subplots(figsize=(5, 5))
    ig.plot(
        g,
        target=ax,
        layout="kk",
        vertex_size=25,
        edge_width=4,
        vertex_label=range(g.vcount()),
        vertex_color="white",
    )



from igraph.drawing.utils import FakeModule
from typing import Any

__all__ = ("find_plotly",)
__docformat__ = "restructuredtext en"


def find_plotly() -> Any:
    """Tries to import the ``plotly`` Python module if it is installed.
    Returns a fake module if everything fails.
    """
    try:
        import plotly

    except ImportError:
        plotly = FakeModule("You need to install plotly to use this functionality")

    return plotly


def format_path_step(code, point_or_points):
    """Format step in SVG path for plotly"""
    if isinstance(point_or_points[0], (float, int)):
        points = [point_or_points]
    else:
        points = point_or_points

    step = f"{code}"
    for point in points:
        x, y = point[0], point[1]
        step += f" {x},{y}"
    return step


# Adapted from https://community.plotly.com/t/arc-shape-with-path/7205/3
def format_arc(center, radius_x, radius_y, theta1, theta2, N=100, closed=False):
    """Approximation of an SVG-style arc

    NOTE: plotly does not currently support the native SVG "A/a" commands"""
    import math

    xc, yc = center
    dt = 1.0 * (theta2 - theta1)
    t = [theta1 + dt * i / (N - 1) for i in range(N)]
    x = [xc + radius_x * math.cos(i) for i in t]
    y = [yc + radius_y * math.sin(i) for i in t]
    path = f"M {x[0]}, {y[0]}"
    for k in range(1, len(t)):
        path += f"L{x[k]}, {y[k]}"
    if closed:
        path += " Z"
    return path


def format_rgba(color):
    """Format colors in a way understood by plotly"""
    if isinstance(color, str):
        return color

    if isinstance(color, float):
        if color > 1:
            color /= 255.0
        color = [color] * 3

    r = int(255 * color[0])
    g = int(255 * color[1])
    b = int(255 * color[2])
    if len(color) > 3:
        a = int(255 * color[3])
    else:
        a = 255

    colstr = f"rgba({r},{g},{b},{a})"
    return colstr



"""This module provides a dendrogram drawer for the Cairo backend."""

from math import pi

from igraph.drawing.utils import str_to_orientation

from .base import AbstractCairoDrawer

__all__ = ("CairoDendrogramDrawer",)


class CairoDendrogramDrawer(AbstractCairoDrawer):
    """Default Cairo drawer object for dendrograms."""

    def __init__(self, context, bbox, palette):
        """Constructs the drawer and associates it to the given palette.

        @param context: the context on which we will draw
        @param bbox:    the bounding box within which we will draw.
                        Can be anything accepted by the constructor
                        of L{BoundingBox} (i.e., a 2-tuple, a 4-tuple
                        or a L{BoundingBox} object).
        @param palette: the palette that can be used to map integer
                        color indices to colors when drawing vertices
        """
        super().__init__(context, bbox)
        self.palette = palette

    @staticmethod
    def _item_box_size(dendro, context, horiz, idx):
        """Calculates the amount of space needed for drawing an
        individual vertex at the bottom of the dendrogram."""
        if dendro._names is None or dendro._names[idx] is None:
            x_bearing, _, _, height, x_advance, _ = context.text_extents("")
        else:
            x_bearing, _, _, height, x_advance, _ = context.text_extents(
                str(dendro._names[idx])
            )

        if horiz:
            return x_advance - x_bearing, height
        return height, x_advance - x_bearing

    def _plot_item(self, dendro, context, horiz, idx, x, y):
        """Plots a dendrogram item to the given Cairo context

        @param context: the Cairo context we are plotting on
        @param horiz: whether the dendrogram is horizontally oriented
        @param idx: the index of the item
        @param x: the X position of the item
        @param y: the Y position of the item
        """
        if dendro._names is None or dendro._names[idx] is None:
            return

        height = self._item_box_size(dendro, context, True, idx)[1]
        if horiz:
            context.move_to(x, y + height)
            context.show_text(str(dendro._names[idx]))
        else:
            context.save()
            context.translate(x, y)
            context.rotate(-pi / 2.0)
            context.move_to(0, height)
            context.show_text(str(dendro._names[idx]))
            context.restore()

    def draw(self, dendro, **kwds):
        """Draws the given Dendrogram in a Cairo context.

        @param dendro: the igraph.Dendrogram to plot.

        It accepts the following keyword arguments:

          - C{style}: the style of the plot. C{boolean} is useful for plotting
            matrices with boolean (C{True}/C{False} or 0/1) values: C{False}
            will be shown with a white box and C{True} with a black box.
            C{palette} uses the given palette to represent numbers by colors,
            the minimum will be assigned to palette color index 0 and the maximum
            will be assigned to the length of the palette. C{None} draws transparent
            cell backgrounds only. The default style is C{boolean} (but it may
            change in the future). C{None} values in the matrix are treated
            specially in both cases: nothing is drawn in the cell corresponding
            to C{None}.

          - C{square}: whether the cells of the matrix should be square or not.
            Default is C{True}.

          - C{grid_width}: line width of the grid shown on the matrix. If zero or
            negative, the grid is turned off. The grid is also turned off if the size
            of a cell is less than three times the given line width. Default is C{1}.
            Fractional widths are also allowed.

          - C{border_width}: line width of the border drawn around the matrix.
            If zero or negative, the border is turned off. Default is C{1}.

          - C{row_names}: the names of the rows

          - C{col_names}: the names of the columns.

          - C{values}: values to be displayed in the cells. If C{None} or
            C{False}, no values are displayed. If C{True}, the values come
            from the matrix being plotted. If it is another matrix, the
            values of that matrix are shown in the cells. In this case,
            the shape of the value matrix must match the shape of the
            matrix being plotted.

          - C{value_format}: a format string or a callable that specifies how
            the values should be plotted. If it is a callable, it must be a
            function that expects a single value and returns a string.
            Example: C{"%#.2f"} for floating-point numbers with always exactly
            two digits after the decimal point. See the Python documentation of
            the C{%} operator for details on the format string. If the format
            string is not given, it defaults to the C{str} function.

        If only the row names or the column names are given and the matrix
        is square-shaped, the same names are used for both column and row
        names.
        """
        from igraph.layout import Layout

        context = self.context
        bbox = self.bbox

        if dendro._names is None:
            dendro._names = [str(x) for x in range(dendro._nitems)]

        orientation = str_to_orientation(
            kwds.get("orientation", "lr"), reversed_vertical=True
        )
        horiz = orientation in ("lr", "rl")

        # Get the font height
        font_height = context.font_extents()[2]

        # Calculate space needed for individual items at the
        # bottom of the dendrogram
        item_boxes = [
            self._item_box_size(dendro, context, horiz, idx)
            for idx in range(dendro._nitems)
        ]

        # Small correction for cases when the right edge of the labels is
        # aligned with the tips of the dendrogram branches
        ygap = 2 if orientation == "bt" else 0
        xgap = 2 if orientation == "lr" else 0
        item_boxes = [(x + xgap, y + ygap) for x, y in item_boxes]

        # Calculate coordinates
        layout = Layout([(0, 0)] * dendro._nitems, dim=2)
        inorder = dendro._traverse_inorder()
        if not horiz:
            x, y = 0, 0
            for element in inorder:
                layout[element] = (x, 0)
                x += max(font_height, item_boxes[element][0])

            for id1, id2 in dendro._merges:
                y += 1
                layout.append(((layout[id1][0] + layout[id2][0]) / 2.0, y))

            # Mirror or rotate the layout if necessary
            if orientation == "bt":
                layout.mirror(1)
        else:
            x, y = 0, 0
            for element in inorder:
                layout[element] = (0, y)
                y += max(font_height, item_boxes[element][1])

            for id1, id2 in dendro._merges:
                x += 1
                layout.append((x, (layout[id1][1] + layout[id2][1]) / 2.0))

            # Mirror or rotate the layout if necessary
            if orientation == "rl":
                layout.mirror(0)

        # Rescale layout to the bounding box
        maxw = max(e[0] for e in item_boxes)
        maxh = max(e[1] for e in item_boxes)

        # w, h: width and height of the area containing the dendrogram
        # tree without the items.
        # delta_x, delta_y: displacement of the dendrogram tree
        width, height = float(bbox.width), float(bbox.height)
        delta_x, delta_y = 0, 0
        if horiz:
            width -= maxw
            if orientation == "lr":
                delta_x = maxw
        else:
            height -= maxh
            if orientation == "tb":
                delta_y = maxh

        if horiz:
            delta_y += font_height / 2.0
        else:
            delta_x += font_height / 2.0
        layout.fit_into(
            (delta_x, delta_y, width - delta_x, height - delta_y),
            keep_aspect_ratio=False,
        )

        context.save()

        context.translate(bbox.left, bbox.top)
        context.set_source_rgb(0.0, 0.0, 0.0)
        context.set_line_width(1)

        # Draw items
        if horiz:
            sgn = 0 if orientation == "rl" else -1
            for idx in range(dendro._nitems):
                x = layout[idx][0] + sgn * item_boxes[idx][0]
                y = layout[idx][1] - item_boxes[idx][1] / 2.0
                self._plot_item(dendro, context, horiz, idx, x, y)
        else:
            sgn = 1 if orientation == "bt" else 0
            for idx in range(dendro._nitems):
                x = layout[idx][0] - item_boxes[idx][0] / 2.0
                y = layout[idx][1] + sgn * item_boxes[idx][1]
                dendro._plot_item(dendro, context, horiz, idx, x, y)

        # Draw dendrogram lines
        if not horiz:
            for idx, (id1, id2) in enumerate(dendro._merges):
                x0, y0 = layout[id1]
                x1, y1 = layout[id2]
                x2, y2 = layout[idx + dendro._nitems]
                context.move_to(x0, y0)
                context.line_to(x0, y2)
                context.line_to(x1, y2)
                context.line_to(x1, y1)
                context.stroke()
        else:
            for idx, (id1, id2) in enumerate(dendro._merges):
                x0, y0 = layout[id1]
                x1, y1 = layout[id2]
                x2, y2 = layout[idx + dendro._nitems]
                context.move_to(x0, y0)
                context.line_to(x2, y0)
                context.line_to(x2, y1)
                context.line_to(x1, y1)
                context.stroke()

        context.restore()



"""
Drawing routines to draw graphs.

This module contains routines to draw graphs on:

  - Cairo surfaces (L{DefaultGraphDrawer})
  - Matplotlib axes (L{MatplotlibGraphDrawer})

It also contains routines to send an igraph graph directly to
(U{Cytoscape<http://www.cytoscape.org>}) using the
(U{CytoscapeRPC plugin<http://gforge.nbic.nl/projects/cytoscaperpc/>}), see
L{CytoscapeGraphDrawer}. L{CytoscapeGraphDrawer} can also fetch the current
network from Cytoscape and convert it to igraph format.
"""

from math import atan2, cos, pi, sin, tan
from warnings import warn

from igraph._igraph import convex_hull, VertexSeq
from igraph.configuration import Configuration
from igraph.drawing.baseclasses import AbstractGraphDrawer
from igraph.drawing.text import TextAlignment
from igraph.drawing.utils import Point

from .base import AbstractCairoDrawer
from .edge import CairoArrowEdgeDrawer
from .polygon import CairoPolygonDrawer
from .text import CairoTextDrawer
from .utils import find_cairo
from .vertex import CairoVertexDrawer

__all__ = ("CairoGraphDrawer",)

cairo = find_cairo()

#####################################################################


#####################################################################


class AbstractCairoGraphDrawer(AbstractGraphDrawer, AbstractCairoDrawer):
    """Abstract base class for graph drawers that draw on a Cairo canvas."""

    def __init__(self, context, bbox):
        """Constructs the graph drawer and associates it to the given
        Cairo context and the given L{BoundingBox}.

        @param context: the context on which we will draw
        @param bbox:    the bounding box within which we will draw.
                        Can be anything accepted by the constructor
                        of L{BoundingBox} (i.e., a 2-tuple, a 4-tuple
                        or a L{BoundingBox} object).
        """
        AbstractCairoDrawer.__init__(self, context, bbox)
        AbstractGraphDrawer.__init__(self)


#####################################################################


class CairoGraphDrawer(AbstractCairoGraphDrawer):
    """Class implementing the default visualisation of a graph.

    The default visualisation of a graph draws the nodes on a 2D plane
    according to a given L{Layout}, then draws a straight or curved
    edge between nodes connected by edges. This is the visualisation
    used when one invokes the L{plot()} function on a L{Graph} object.

    See L{Graph.__plot__()} for the keyword arguments understood by
    this drawer."""

    def __init__(
        self,
        context,
        bbox=None,
        vertex_drawer_factory=CairoVertexDrawer,
        edge_drawer_factory=CairoArrowEdgeDrawer,
        label_drawer_factory=CairoTextDrawer,
    ):
        """Constructs the graph drawer and associates it to the given
        Cairo context and the given L{BoundingBox}.

        @param context: the context on which we will draw
        @param bbox:    the bounding box within which we will draw.
                        Can be anything accepted by the constructor
                        of L{BoundingBox} (i.e., a 2-tuple, a 4-tuple
                        or a L{BoundingBox} object).
        @param vertex_drawer_factory: a factory method that returns an
                        L{AbstractCairoVertexDrawer} instance bound to a
                        given Cairo context. The factory method must take
                        four parameters: the Cairo context, the bounding
                        box of the drawing area, the palette to be
                        used for drawing colored vertices, and the graph layout.
                        The default vertex drawer is L{CairoVertexDrawer}.
        @param edge_drawer_factory: a factory method that returns an
                        L{AbstractCairoEdgeDrawer} instance bound to a
                        given Cairo context. The factory method must take
                        two parameters: the Cairo context and the palette
                        to be used for drawing colored edges. You can use
                        any of the actual L{AbstractEdgeDrawer}
                        implementations here to control the style of
                        edges drawn by igraph. The default edge drawer is
                        L{CairoArrowEdgeDrawer}.
        @param label_drawer_factory: a factory method that returns a
                        L{CairoTextDrawer} instance bound to a given Cairo
                        context. The method must take one parameter: the
                        Cairo context. The default label drawer is
                        L{CairoTextDrawer}.
        """
        super().__init__(context, bbox)
        self.vertex_drawer_factory = vertex_drawer_factory
        self.edge_drawer_factory = edge_drawer_factory
        self.label_drawer_factory = label_drawer_factory

    def draw(self, graph, *args, **kwds):
        # Positional arguments are not used
        if args:
            warn(
                "Positional arguments to plot functions are ignored "
                "and will be deprecated soon.",
                DeprecationWarning,
                stacklevel=1,
            )

        bbox = kwds.pop("bbox", None)
        if bbox is None:
            raise ValueError("bbox is required for Cairo plots")
        # Validate it through set/get
        self.bbox = bbox
        bbox = self.bbox

        # Some abbreviations for sake of simplicity
        directed = graph.is_directed()
        context = self.context

        # Palette
        palette = kwds.pop("palette", None)

        # Calculate/get the layout of the graph
        layout = self.ensure_layout(kwds.get("layout", None), graph)

        # Determine the size of the margin on each side
        margin = kwds.get("margin", 0)
        try:
            margin = list(margin)  # type: ignore
        except TypeError:
            margin = [margin]
        while len(margin) < 4:
            margin.extend(margin)

        # Contract the drawing area by the margin and fit the layout
        bbox = self.bbox.contract(margin)
        layout.fit_into(bbox, keep_aspect_ratio=kwds.get("keep_aspect_ratio", False))

        # Decide whether we need to calculate the curvature of edges
        # automatically -- and calculate them if needed.
        autocurve = kwds.get("autocurve", None)
        if autocurve or (
            autocurve is None
            and "edge_curved" not in kwds
            and "curved" not in graph.edge_attributes()
            and graph.ecount() < 10000
        ):
            from igraph import autocurve

            default = kwds.get("edge_curved", 0)
            if default is True:
                default = 0.5
            default = float(default)
            kwds["edge_curved"] = autocurve(graph, attribute=None, default=default)

        # Construct the vertex, edge and label drawers
        vertex_drawer = self.vertex_drawer_factory(context, bbox, palette, layout)
        edge_drawer = self.edge_drawer_factory(context, palette)
        label_drawer = self.label_drawer_factory(context)

        # Construct the visual vertex/edge builders based on the specifications
        # provided by the vertex_drawer and the edge_drawer
        vertex_builder = vertex_drawer.VisualVertexBuilder(graph.vs, kwds)
        edge_builder = edge_drawer.VisualEdgeBuilder(graph.es, kwds)

        # Determine the order in which we will draw the vertices and edges
        vertex_order = self._determine_vertex_order(graph, kwds)
        edge_order = self._determine_edge_order(graph, kwds)

        # Draw the highlighted groups (if any)
        if "mark_groups" in kwds:
            mark_groups = kwds["mark_groups"]

            # Deferred import to avoid a cycle in the import graph
            from igraph.clustering import VertexClustering, VertexCover

            # Figure out what to do with mark_groups in order to be able to
            # iterate over it and get memberlist-color pairs
            if isinstance(mark_groups, dict):
                # Dictionary mapping vertex indices or tuples of vertex
                # indices to colors
                group_iter = iter(mark_groups.items())
            elif isinstance(mark_groups, (VertexClustering, VertexCover)):
                # Vertex clustering
                group_iter = ((group, color) for color, group in enumerate(mark_groups))
            elif hasattr(mark_groups, "__iter__"):
                # Lists, tuples, iterators etc
                group_iter = iter(mark_groups)
            else:
                # False
                group_iter = iter({}.items())

            # We will need a polygon drawer to draw the convex hulls
            polygon_drawer = CairoPolygonDrawer(context, bbox)

            # Iterate over color-memberlist pairs
            for group, color_id in group_iter:
                if not group or color_id is None:
                    continue

                color = palette.get(color_id)

                if isinstance(group, VertexSeq):
                    group = [vertex.index for vertex in group]
                if not hasattr(group, "__iter__"):
                    raise TypeError("group membership list must be iterable")

                # Get the vertex indices that constitute the convex hull
                hull = [group[i] for i in convex_hull([layout[idx] for idx in group])]

                # Calculate the preferred rounding radius for the corners
                corner_radius = 1.25 * max(vertex_builder[idx].size for idx in hull)

                # Construct the polygon
                polygon = [layout[idx] for idx in hull]

                if len(polygon) == 2:
                    # Expand the polygon (which is a flat line otherwise)
                    a, b = Point(*polygon[0]), Point(*polygon[1])
                    c = corner_radius * (a - b).normalized()
                    n = Point(-c[1], c[0])
                    polygon = [a + n, b + n, b - c, b - n, a - n, a + c]
                else:
                    # Expand the polygon around its center of mass
                    center = Point(
                        *[sum(coords) / float(len(coords)) for coords in zip(*polygon)]
                    )
                    polygon = [
                        Point(*point).towards(center, -corner_radius)
                        for point in polygon
                    ]

                # Draw the hull
                context.set_source_rgba(color[0], color[1], color[2], color[3] * 0.25)
                polygon_drawer.draw_path(polygon, corner_radius=corner_radius)
                context.fill_preserve()
                context.set_source_rgba(*color)
                context.stroke()

        # Construct the iterator that we will use to draw the edges
        es = graph.es
        if edge_order is None:
            # Default edge order
            edge_coord_iter = zip(es, edge_builder)
        else:
            # Specified edge order
            edge_coord_iter = ((es[i], edge_builder[i]) for i in edge_order)

        # Draw the edges
        if directed:
            drawer_method = edge_drawer.draw_directed_edge
        else:
            drawer_method = edge_drawer.draw_undirected_edge
        for edge, visual_edge in edge_coord_iter:
            src, dest = edge.tuple
            src_vertex, dest_vertex = vertex_builder[src], vertex_builder[dest]
            drawer_method(visual_edge, src_vertex, dest_vertex)

        # Construct the iterator that we will use to draw the vertices
        vs = graph.vs
        if vertex_order is None:
            # Default vertex order
            vertex_coord_iter = zip(vs, vertex_builder, layout)
        else:
            # Specified vertex order
            vertex_coord_iter = (
                (vs[i], vertex_builder[i], layout[i]) for i in vertex_order
            )

        # Draw the vertices
        drawer_method = vertex_drawer.draw
        context.set_line_width(1)
        for vertex, visual_vertex, coords in vertex_coord_iter:
            drawer_method(visual_vertex, vertex, coords)

        # Decide whether the labels have to be wrapped
        wrap = kwds.get("wrap_labels")
        if wrap is None:
            wrap = Configuration.instance()["plotting.wrap_labels"]
        wrap = bool(wrap)

        # Construct the iterator that we will use to draw the vertex labels
        if vertex_order is None:
            # Default vertex order
            vertex_coord_iter = zip(vertex_builder, layout)
        else:
            # Specified vertex order
            vertex_coord_iter = ((vertex_builder[i], layout[i]) for i in vertex_order)

        # Draw the vertex labels
        for vertex, coords in vertex_coord_iter:
            if vertex.label is None:
                continue

            # Set the font family, size, color and text
            context.select_font_face(
                vertex.font, cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL
            )
            context.set_font_size(vertex.label_size)
            context.set_source_rgba(*vertex.label_color)
            label_drawer.text = vertex.label

            if vertex.label_dist:
                # Label is displaced from the center of the vertex.
                _, yb, w, h, _, _ = label_drawer.text_extents()
                w, h = w / 2.0, h / 2.0
                radius = vertex.label_dist * vertex.size / 2.0
                # First we find the reference point that is at distance `radius'
                # from the vertex in the direction given by `label_angle'.
                # Then we place the label in a way that the line connecting the
                # center of the bounding box of the label with the center of the
                # vertex goes through the reference point and the reference
                # point lies exactly on the bounding box of the vertex.
                alpha = vertex.label_angle % (2 * pi)
                cx = coords[0] + radius * cos(alpha)
                cy = coords[1] - radius * sin(alpha)
                # Now we have the reference point. We have to decide which side
                # of the label box will intersect with the line that connects
                # the center of the label with the center of the vertex.
                if w > 0:
                    beta = atan2(h, w) % (2 * pi)
                else:
                    beta = pi / 2.0
                gamma = pi - beta
                if alpha > 2 * pi - beta or alpha <= beta:
                    # Intersection at left edge of label
                    cx += w
                    cy -= tan(alpha) * w
                elif alpha > beta and alpha <= gamma:
                    # Intersection at bottom edge of label
                    try:
                        cx += h / tan(alpha)
                    except Exception:
                        pass  # tan(alpha) == inf
                    cy -= h
                elif alpha > gamma and alpha <= gamma + 2 * beta:
                    # Intersection at right edge of label
                    cx -= w
                    cy += tan(alpha) * w
                else:
                    # Intersection at top edge of label
                    try:
                        cx -= h / tan(alpha)
                    except Exception:
                        pass  # tan(alpha) == inf
                    cy += h
                # Draw the label
                label_drawer.draw_at(cx - w, cy - h - yb, wrap=wrap)
            else:
                # Label is exactly in the center of the vertex
                cx, cy = coords
                half_size = vertex.size / 2.0
                label_drawer.bbox = (
                    cx - half_size,
                    cy - half_size,
                    cx + half_size,
                    cy + half_size,
                )
                label_drawer.draw(wrap=wrap)

        # Construct the iterator that we will use to draw the edge labels
        es = graph.es
        if edge_order is None:
            # Default edge order
            edge_coord_iter = zip(es, edge_builder)
        else:
            # Specified edge order
            edge_coord_iter = ((es[i], edge_builder[i]) for i in edge_order)

        # Draw the edge labels
        for edge, visual_edge in edge_coord_iter:
            if visual_edge.label is None:
                continue

            # Set the font family, size, color and text
            context.select_font_face(
                visual_edge.font, cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL
            )
            context.set_font_size(visual_edge.label_size)
            context.set_source_rgba(*visual_edge.label_color)
            label_drawer.text = visual_edge.label

            # Ask the edge drawer to propose an anchor point for the label
            src, dest = edge.tuple
            src_vertex, dest_vertex = vertex_builder[src], vertex_builder[dest]
            (x, y), (halign, valign) = edge_drawer.get_label_position(
                visual_edge, src_vertex, dest_vertex
            )

            # Measure the text
            _, yb, w, h, _, _ = label_drawer.text_extents()
            w /= 2.0
            h /= 2.0

            # Place the text relative to the edge
            if halign == TextAlignment.RIGHT:
                x -= w
            elif halign == TextAlignment.LEFT:
                x += w
            if valign == TextAlignment.BOTTOM:
                y -= h - yb / 2.0
            elif valign == TextAlignment.TOP:
                y += h

            # Draw the edge label
            label_drawer.halign = halign
            label_drawer.valign = valign
            label_drawer.bbox = (x - w, y - h, x + w, y + h)
            label_drawer.draw(wrap=wrap)



import random
import unittest

from igraph import Graph, Matrix


class DirectedUndirectedTests(unittest.TestCase):
    def testToUndirected(self):
        graph = Graph([(0, 1), (0, 2), (1, 0)], directed=True)

        graph2 = graph.copy()
        graph2.to_undirected(mode=False)
        self.assertTrue(graph2.vcount() == graph.vcount())
        self.assertTrue(graph2.is_directed() is False)
        self.assertTrue(sorted(graph2.get_edgelist()) == [(0, 1), (0, 1), (0, 2)])

        graph2 = graph.copy()
        graph2.to_undirected()
        self.assertTrue(graph2.vcount() == graph.vcount())
        self.assertTrue(graph2.is_directed() is False)
        self.assertTrue(sorted(graph2.get_edgelist()) == [(0, 1), (0, 2)])

        graph2 = graph.copy()
        graph2.es["weight"] = [1, 2, 3]
        graph2.to_undirected(mode="collapse", combine_edges="sum")
        self.assertTrue(graph2.vcount() == graph.vcount())
        self.assertTrue(graph2.is_directed() is False)
        self.assertTrue(sorted(graph2.get_edgelist()) == [(0, 1), (0, 2)])
        self.assertTrue(graph2.es["weight"] == [4, 2])

        graph = Graph([(0, 1), (1, 0), (0, 1), (1, 0), (2, 1), (1, 2)], directed=True)
        graph2 = graph.copy()
        graph2.es["weight"] = [1, 2, 3, 4, 5, 6]
        graph2.to_undirected(mode="mutual", combine_edges="sum")
        self.assertTrue(graph2.vcount() == graph.vcount())
        self.assertTrue(graph2.is_directed() is False)
        self.assertTrue(sorted(graph2.get_edgelist()) == [(0, 1), (0, 1), (1, 2)])
        self.assertTrue(
            graph2.es["weight"] == [7, 3, 11] or graph2.es["weight"] == [3, 7, 11]
        )

    def testToDirectedNoModeArg(self):
        graph = Graph([(0, 1), (0, 2), (2, 3), (2, 4)], directed=False)
        graph.to_directed()
        self.assertTrue(graph.is_directed())
        self.assertTrue(graph.vcount() == 5)
        self.assertTrue(
            sorted(graph.get_edgelist())
            == [(0, 1), (0, 2), (1, 0), (2, 0), (2, 3), (2, 4), (3, 2), (4, 2)]
        )

    def testToDirectedMutual(self):
        graph = Graph([(0, 1), (0, 2), (2, 3), (2, 4)], directed=False)
        graph.to_directed("mutual")
        self.assertTrue(graph.is_directed())
        self.assertTrue(graph.vcount() == 5)
        self.assertTrue(
            sorted(graph.get_edgelist())
            == [(0, 1), (0, 2), (1, 0), (2, 0), (2, 3), (2, 4), (3, 2), (4, 2)]
        )

    def testToDirectedAcyclic(self):
        graph = Graph([(0, 1), (2, 0), (3, 0), (3, 0), (4, 2)], directed=False)
        graph.to_directed("acyclic")
        self.assertTrue(graph.is_directed())
        self.assertTrue(graph.vcount() == 5)
        self.assertTrue(
            sorted(graph.get_edgelist()) == [(0, 1), (0, 2), (0, 3), (0, 3), (2, 4)]
        )

    def testToDirectedRandom(self):
        random.seed(0)

        graph = Graph.Ring(200, directed=False)
        graph.to_directed("random")

        self.assertTrue(graph.is_directed())
        self.assertTrue(graph.vcount() == 200)
        edgelist1 = sorted(graph.get_edgelist())

        graph = Graph.Ring(200, directed=False)
        graph.to_directed("random")

        self.assertTrue(graph.is_directed())
        self.assertTrue(graph.vcount() == 200)
        edgelist2 = sorted(graph.get_edgelist())

        self.assertTrue(edgelist1 != edgelist2)

    def testToDirectedInvalidMode(self):
        graph = Graph([(0, 1), (0, 2), (2, 3), (2, 4)], directed=False)
        with self.assertRaises(ValueError):
            graph.to_directed("no-such-mode")


class GraphRepresentationTests(unittest.TestCase):
    def testGetAdjacency(self):
        # Undirected case
        g = Graph.Tree(6, 3)
        g.es["weight"] = list(range(5))
        self.assertTrue(
            g.get_adjacency()
            == Matrix(
                [
                    [0, 1, 1, 1, 0, 0],
                    [1, 0, 0, 0, 1, 1],
                    [1, 0, 0, 0, 0, 0],
                    [1, 0, 0, 0, 0, 0],
                    [0, 1, 0, 0, 0, 0],
                    [0, 1, 0, 0, 0, 0],
                ]
            )
        )
        self.assertTrue(
            g.get_adjacency(attribute="weight")
            == Matrix(
                [
                    [0, 0, 1, 2, 0, 0],
                    [0, 0, 0, 0, 3, 4],
                    [1, 0, 0, 0, 0, 0],
                    [2, 0, 0, 0, 0, 0],
                    [0, 3, 0, 0, 0, 0],
                    [0, 4, 0, 0, 0, 0],
                ]
            )
        )

        # Directed case
        g = Graph.Tree(6, 3, "tree_out")
        g.add_edges([(0, 1), (1, 0)])
        self.assertTrue(
            g.get_adjacency()
            == Matrix(
                [
                    [0, 2, 1, 1, 0, 0],
                    [1, 0, 0, 0, 1, 1],
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                ]
            )
        )

    def testGetSparseAdjacency(self):
        try:
            from scipy import sparse  # noqa: F401
            import numpy as np
        except ImportError:
            self.skipTest("Scipy and numpy are dependencies of this test.")

        # Undirected case
        g = Graph.Tree(6, 3)
        g.es["weight"] = list(range(5))
        self.assertTrue(
            np.all((g.get_adjacency_sparse() == np.array(g.get_adjacency().data)))
        )
        self.assertTrue(
            np.all(
                (
                    g.get_adjacency_sparse(attribute="weight")
                    == np.array(g.get_adjacency(attribute="weight").data)
                )
            )
        )

        # Directed case
        g = Graph.Tree(6, 3, "tree_out")
        g.add_edges([(0, 1), (1, 0)])
        self.assertTrue(
            np.all(g.get_adjacency_sparse() == np.array(g.get_adjacency().data))
        )

    def testGetAdjacencyRoundtrip(self):
        g = Graph.Tree(6, 3)
        adj = g.get_adjacency()
        g2 = Graph.Adjacency(adj, mode="undirected")
        self.assertEqual(g.vcount(), g2.vcount())
        self.assertEqual(g.is_directed(), g2.is_directed())
        self.assertTrue(g.get_edgelist() == g2.get_edgelist())


class PruferTests(unittest.TestCase):
    def testFromPrufer(self):
        g = Graph.Prufer([3, 3, 3, 4])
        self.assertEqual(6, g.vcount())
        self.assertEqual(5, g.ecount())
        self.assertEqual(
            [(0, 3), (1, 3), (2, 3), (3, 4), (4, 5)], sorted(g.get_edgelist())
        )

    def testToPrufer(self):
        g = Graph([(0, 3), (1, 3), (2, 3), (3, 4), (4, 5)])
        self.assertEqual([3, 3, 3, 4], g.to_prufer())


def suite():
    direction_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        DirectedUndirectedTests
    )
    representation_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        GraphRepresentationTests
    )
    prufer_suite = unittest.defaultTestLoader.loadTestsFromTestCase(PruferTests)
    return unittest.TestSuite([direction_suite, representation_suite, prufer_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



"""This module provides implementation for a Matplotlib-specific histogram drawer."""

from igraph.drawing.baseclasses import AbstractDrawer

__all__ = ("MatplotlibHistogramDrawer",)


class MatplotlibHistogramDrawer(AbstractDrawer):
    """Matplotlib drawer object for matrices."""

    def __init__(self, ax):
        """Constructs the drawer and associates it to the given Axes.

        @param ax: the Axes on which we will draw
        """
        self.context = ax

    def draw(self, matrix, **kwds):
        """Draws the given Matrix in a matplotlib Axes.

        @param matrix: the igraph.Histogram to plot.

        """
        ax = self.context

        xmin = kwds.get("min", self._min)
        ymin = 0
        xmax = kwds.get("max", self._max)
        ymax = kwds.get("max_value", max(self._bins))
        width = self._bin_width

        x = [self._min + width * i for i, _ in enumerate(self._bins)]
        y = self._bins
        # Draw the boxes/bars
        ax.bar(x, y, align="left")
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)



"""
.. _tutorials-maxflow:

============
Maximum Flow
============

This example shows how to construct a max flow on a directed graph with edge capacities using :meth:`igraph.Graph.maxflow`.

"""

import igraph as ig
import matplotlib.pyplot as plt

# %%
# First, we generate a graph and assign a "capacity" to each edge:
g = ig.Graph(6, [(3, 2), (3, 4), (2, 1), (4, 1), (4, 5), (1, 0), (5, 0)], directed=True)
g.es["capacity"] = [7, 8, 1, 2, 3, 4, 5]

# %%
# To find the max flow, we can simply run:
flow = g.maxflow(3, 0, capacity=g.es["capacity"])

print("Max flow:", flow.value)
print("Edge assignments:", flow.flow)

# Output:
# Max flow: 6.0
# Edge assignments [1.0, 5.0, 1.0, 2.0, 3.0, 3.0, 3.0]

# %%
# Finally, we can plot the directed graph to look at the situation:
fig, ax = plt.subplots()
ig.plot(
    g,
    target=ax,
    layout="circle",
    vertex_label=range(g.vcount()),
    vertex_color="lightblue",
)
plt.show()



from igraph.drawing.utils import calculate_corner_radii
from igraph.utils import consecutive_pairs

from .base import AbstractCairoDrawer

__all__ = ("CairoPolygonDrawer",)


class CairoPolygonDrawer(AbstractCairoDrawer):
    """Class that is used to draw polygons in Cairo.

    The corner points of the polygon can be set by the C{points}
    property of the drawer, or passed at construction time. Most
    drawing methods in this class also have an extra C{points}
    argument that can be used to override the set of points in the
    C{points} property."""

    def __init__(self, context, bbox=(1, 1)):
        """Constructs a new polygon drawer that draws on the given
        Cairo context.

        @param  context: the Cairo context to draw on
        @param  bbox:    ignored, leave it at its default value
        """
        super().__init__(context, bbox)

    def draw_path(self, points, corner_radius=0):
        """Sets up a Cairo path for the outline of a polygon on the given
        Cairo context.

        @param points: the coordinates of the corners of the polygon,
          in clockwise or counter-clockwise order.
        @param corner_radius: if zero, an ordinary polygon will be drawn.
          If positive, the corners of the polygon will be rounded with
          the given radius.
        """
        self.context.new_path()

        if len(points) < 2:
            # Well, a polygon must have at least two corner points
            return

        ctx = self.context
        if corner_radius <= 0:
            # No rounded corners, this is simple
            ctx.move_to(*points[-1])
            for point in points:
                ctx.line_to(*point)
            return

        # Rounded corners. First, we will take each side of the
        # polygon and find what the corner radius should be on
        # each corner. If the side is longer than 2r (where r is
        # equal to corner_radius), the radius allowed by that side
        # is r; if the side is shorter, the radius is the length
        # of the side / 2. For each corner, the final corner radius
        # is the smaller of the radii on the two sides adjacent to
        # the corner.
        corner_radii = calculate_corner_radii(points, corner_radius)

        # Okay, move to the last corner, adjusted by corner_radii[-1]
        # towards the first corner
        ctx.move_to(*(points[-1].towards(points[0], corner_radii[-1])))
        # Now, for each point in points, draw a line towards the
        # corner, stopping before it in a distance of corner_radii[idx],
        # then draw the corner
        u = points[-1]
        for idx, (v, w) in enumerate(consecutive_pairs(points, True)):
            radius = corner_radii[idx]
            ctx.line_to(*v.towards(u, radius))
            aux1 = v.towards(u, radius / 2)
            aux2 = v.towards(w, radius / 2)
            ctx.curve_to(
                aux1.x, aux1.y, aux2.x, aux2.y, *v.towards(w, corner_radii[idx])
            )
            u = v

    def draw(self, points):
        """Draws the polygon using the current stroke of the Cairo context.

        @param points: the coordinates of the corners of the polygon,
          in clockwise or counter-clockwise order.
        """
        self.draw_path(points)
        self.context.stroke()



"""Drawers for various edge styles in Matplotlib graph plots."""

from math import atan2, cos, pi, sin

from igraph.drawing.baseclasses import AbstractEdgeDrawer
from igraph.drawing.metamagic import AttributeCollectorBase
from igraph.drawing.plotly.utils import (
    find_plotly,
    format_path_step,
    format_arc,
    format_rgba,
)
from igraph.drawing.utils import (
    Point,
    euclidean_distance,
    get_bezier_control_points_for_curved_edge,
    intersect_bezier_curve_and_circle,
)

__all__ = ("PlotlyEdgeDrawer",)

plotly = find_plotly()


class PlotlyEdgeDrawer(AbstractEdgeDrawer):
    """Matplotlib-specific abstract edge drawer object."""

    def __init__(self, context, palette):
        """Constructs the edge drawer.

        @param context: a plotly Figure object on which the edges will be
            drawn.
        @param palette: the palette that can be used to map integer color
            indices to colors when drawing edges
        """
        self.context = context
        self.palette = palette
        self.VisualEdgeBuilder = self._construct_visual_edge_builder()

    def _construct_visual_edge_builder(self):
        """Construct the visual edge builder that will collect the visual
        attributes of an edge when it is being drawn."""

        class VisualEdgeBuilder(AttributeCollectorBase):
            """Builder that collects some visual properties of an edge for
            drawing"""

            _kwds_prefix = "edge_"
            arrow_size = 0.007
            arrow_width = 1.4
            color = "#444"
            curved = (0.0, self._curvature_to_float)
            label = None
            label_color = ("black", self.palette.get)
            label_size = 12.0
            font = "sans-serif"
            width = 2.0

        return VisualEdgeBuilder

    def draw_directed_edge(self, edge, src_vertex, dest_vertex):
        if src_vertex == dest_vertex:  # TODO
            return self.draw_loop_edge(edge, src_vertex)

        fig = self.context
        (x1, y1), (x2, y2) = src_vertex.position, dest_vertex.position
        (x_src, y_src), (x_dest, y_dest) = src_vertex.position, dest_vertex.position

        # Draw the edge
        path = [
            format_path_step("M", [x1, y1]),
        ]

        if edge.curved:
            # Calculate the curve
            aux1, aux2 = get_bezier_control_points_for_curved_edge(
                x1, y1, x2, y2, edge.curved
            )

            # Coordinates of the control points of the Bezier curve
            xc1, yc1 = aux1
            xc2, yc2 = aux2

            # Determine where the edge intersects the circumference of the
            # vertex shape: Tip of the arrow
            x2, y2 = intersect_bezier_curve_and_circle(
                x_src, y_src, xc1, yc1, xc2, yc2, x_dest, y_dest, dest_vertex.size / 2.0
            )

            # Calculate the arrow head coordinates
            angle = atan2(y_dest - y2, x_dest - x2)  # navid
            arrow_size = 15.0 * edge.arrow_size
            arrow_width = 10.0 / edge.arrow_width
            aux_points = [
                (
                    x2 - arrow_size * cos(angle - pi / arrow_width),
                    y2 - arrow_size * sin(angle - pi / arrow_width),
                ),
                (
                    x2 - arrow_size * cos(angle + pi / arrow_width),
                    y2 - arrow_size * sin(angle + pi / arrow_width),
                ),
            ]

            # Midpoint of the base of the arrow triangle
            x_arrow_mid, y_arrow_mid = (
                (aux_points[0][0] + aux_points[1][0]) / 2.0,
                (aux_points[0][1] + aux_points[1][1]) / 2.0,
            )

            # Vector representing the base of the arrow triangle
            x_arrow_base_vec, y_arrow_base_vec = (
                (aux_points[0][0] - aux_points[1][0]),
                (aux_points[0][1] - aux_points[1][1]),
            )

            # Recalculate the curve such that it lands on the base of the arrow triangle
            aux1, aux2 = get_bezier_control_points_for_curved_edge(
                x_src, y_src, x_arrow_mid, y_arrow_mid, edge.curved
            )

            # Offset the second control point (aux2) such that it falls precisely
            # on the normal to the arrow base vector. Strictly speaking,
            # offset_length is the offset length divided by the length of the
            # arrow base vector.
            offset_length = (x_arrow_mid - aux2[0]) * x_arrow_base_vec + (
                y_arrow_mid - aux2[1]
            ) * y_arrow_base_vec
            offset_length /= (
                euclidean_distance(0, 0, x_arrow_base_vec, y_arrow_base_vec) ** 2
            )

            aux2 = (
                aux2[0] + x_arrow_base_vec * offset_length,
                aux2[1] + y_arrow_base_vec * offset_length,
            )

            # Draw the curve from the first vertex to the midpoint of the base
            # of the arrow head
            path.append(format_path_step("C", [aux1, aux2, [x_arrow_mid, y_arrow_mid]]))

        else:
            # FIXME: this is tricky in plotly, let's skip for now
            ## Determine where the edge intersects the circumference of the
            ## vertex shape.
            # x2, y2 = dest_vertex.shape.intersection_point(
            #    x2, y2, x1, y1, dest_vertex.size
            # )

            # Draw the arrowhead
            angle = atan2(y_dest - y2, x_dest - x2)
            arrow_size = 15.0 * edge.arrow_size
            arrow_width = 10.0 / edge.arrow_width
            aux_points = [
                (
                    x2 - arrow_size * cos(angle - pi / arrow_width),
                    y2 - arrow_size * sin(angle - pi / arrow_width),
                ),
                (
                    x2 - arrow_size * cos(angle + pi / arrow_width),
                    y2 - arrow_size * sin(angle + pi / arrow_width),
                ),
            ]

            # Midpoint of the base of the arrow triangle
            x_arrow_mid, y_arrow_mid = (
                (aux_points[0][0] + aux_points[1][0]) / 2.0,
                (aux_points[0][1] + aux_points[1][1]) / 2.0,
            )
            # Draw the line
            path.append(
                format_path_step(
                    "L",
                    Point(x_arrow_mid, y_arrow_mid),
                )
            )

        path = " ".join(path)

        # Draw the edge
        stroke = {
            "type": "path",
            "path": path,
            "line_color": format_rgba(edge.color),
            "line_width": edge.width,
        }
        fig.add_shape(stroke)

        # Draw the arrow head
        arrowhead = plotly.graph_objects.Scatter(
            x=[x2, aux_points[0][0], aux_points[1][0], x2],
            y=[y2, aux_points[0][1], aux_points[1][1], y2],
            fillcolor=format_rgba(edge.color),
            mode="lines",
        )
        fig.add_trace(arrowhead)

    def draw_loop_edge(self, edge, vertex):
        """Draws a loop edge.

        The default implementation draws a small circle.

        @param edge: the edge to be drawn. Visual properties of the edge
          are defined by the attributes of this object.
        @param vertex: the vertex to which the edge is attached. Visual
          properties are given again as attributes.
        """
        fig = self.context
        radius = vertex.size * 1.5
        center_x = vertex.position[0] + cos(pi / 4) * radius / 2.0
        center_y = vertex.position[1] - sin(pi / 4) * radius / 2.0
        stroke = {
            "type": "path",
            "path": format_arc(
                (center_x, center_y),
                radius / 2.0,
                radius / 2.0,
                theta1=0,
                theta2=360.0,
            ),
            "line_color": edge.color,
            "line_width": edge.width,
        }
        fig.add_shape(stroke)

    def draw_undirected_edge(self, edge, src_vertex, dest_vertex):
        """Draws an undirected edge.

        The default implementation of this method draws undirected edges
        as straight lines. Loop edges are drawn as small circles.

        @param edge: the edge to be drawn. Visual properties of the edge
          are defined by the attributes of this object.
        @param src_vertex: the source vertex. Visual properties are given
          again as attributes.
        @param dest_vertex: the target vertex. Visual properties are given
          again as attributes.
        """
        if src_vertex == dest_vertex:
            return self.draw_loop_edge(edge, src_vertex)

        fig = self.context

        path = [format_path_step("M", src_vertex.position)]

        if edge.curved:
            (x1, y1), (x2, y2) = src_vertex.position, dest_vertex.position
            aux1, aux2 = get_bezier_control_points_for_curved_edge(
                x1, y1, x2, y2, edge.curved
            )

            path.append(
                format_path_step(
                    "C",
                    [aux1, aux2, dest_vertex.position],
                )
            )

        else:
            path.append(
                format_path_step(
                    "L",
                    dest_vertex.position,
                )
            )

        path = " ".join(path)

        stroke = {
            "type": "path",
            "path": path,
            "line_color": format_rgba(edge.color),
            "line_width": edge.width,
        }
        fig.add_shape(stroke, layer="below")



import random
import unittest

from igraph import Graph, set_random_number_generator


class FakeRNG:
    @staticmethod
    def random():
        return 0.1

    @staticmethod
    def randint(a, b):
        return a

    @staticmethod
    def gauss(mu, sigma):
        return 0.3


class InvalidRNG:
    pass


class RandomNumberGeneratorTests(unittest.TestCase):
    def tearDown(self):
        set_random_number_generator(random)

    def testSetRandomNumberGenerator(self):
        set_random_number_generator(FakeRNG)
        graph = Graph.GRG(10, 0.2)
        self.assertEqual(graph.vs["x"], [0.1] * 10)
        self.assertEqual(graph.vs["y"], [0.1] * 10)

        self.assertRaises(AttributeError, set_random_number_generator, InvalidRNG)

    def testSeeding(self):
        state = random.getstate()
        g1 = Graph.Erdos_Renyi(n=1000, m=5000)
        random.setstate(state)
        g2 = Graph.Erdos_Renyi(n=1000, m=5000)
        self.assertTrue(g1.get_edgelist() == g2.get_edgelist())


def suite():
    random_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        RandomNumberGeneratorTests
    )
    return unittest.TestSuite([random_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



from igraph._igraph import (
    GET_ADJACENCY_BOTH,
    GET_ADJACENCY_LOWER,
    GET_ADJACENCY_UPPER,
    GraphBase,
)
from igraph.datatypes import Matrix


__all__ = (
    "_get_adjacency",
    "_get_adjacency_sparse",
    "_get_adjlist",
    "_get_biadjacency",
    "_get_inclist",
)


def _get_adjacency(
    self, type=GET_ADJACENCY_BOTH, attribute=None, default=0, eids=False
):
    """Returns the adjacency matrix of a graph.

    @param type: either C{GET_ADJACENCY_LOWER} (uses the lower
      triangle of the matrix) or C{GET_ADJACENCY_UPPER}
      (uses the upper triangle) or C{GET_ADJACENCY_BOTH}
      (uses both parts). Ignored for directed graphs.
    @param attribute: if C{None}, returns the ordinary adjacency
      matrix. When the name of a valid edge attribute is given
      here, the matrix returned will contain the default value
      at the places where there is no edge or the value of the
      given attribute where there is an edge. Multiple edges are
      not supported, the value written in the matrix in this case
      will be unpredictable. This parameter is ignored if
      I{eids} is C{True}
    @param default: the default value written to the cells in the
      case of adjacency matrices with attributes.
    @param eids: specifies whether the edge IDs should be returned
      in the adjacency matrix. Since zero is a valid edge ID, the
      cells in the matrix that correspond to unconnected vertex
      pairs will contain -1 instead of 0 if I{eids} is C{True}.
      If I{eids} is C{False}, the number of edges will be returned
      in the matrix for each vertex pair.
    @return: the adjacency matrix as a L{Matrix}.
    """
    if (
        type != GET_ADJACENCY_LOWER
        and type != GET_ADJACENCY_UPPER
        and type != GET_ADJACENCY_BOTH
    ):
        # Maybe it was called with the first argument as the attribute name
        type, attribute = attribute, type
        if type is None:
            type = GET_ADJACENCY_BOTH

    if eids:
        result = Matrix(GraphBase.get_adjacency(self, type, eids))
        result -= 1
        return result

    if attribute is None:
        return Matrix(GraphBase.get_adjacency(self, type))

    if attribute not in self.es.attribute_names():
        raise ValueError("Attribute does not exist")

    data = [[default] * self.vcount() for _ in range(self.vcount())]

    if self.is_directed():
        for edge in self.es:
            data[edge.source][edge.target] = edge[attribute]
        return Matrix(data)

    if type == GET_ADJACENCY_BOTH:
        for edge in self.es:
            source, target = edge.tuple
            data[source][target] = edge[attribute]
            data[target][source] = edge[attribute]
    elif type == GET_ADJACENCY_UPPER:
        for edge in self.es:
            data[min(edge.tuple)][max(edge.tuple)] = edge[attribute]
    else:
        for edge in self.es:
            data[max(edge.tuple)][min(edge.tuple)] = edge[attribute]

    return Matrix(data)


def _get_adjacency_sparse(self, attribute=None):
    """Returns the adjacency matrix of a graph as a SciPy CSR matrix.

    @param attribute: if C{None}, returns the ordinary adjacency
      matrix. When the name of a valid edge attribute is given
      here, the matrix returned will contain the default value
      at the places where there is no edge or the value of the
      given attribute where there is an edge.
    @return: the adjacency matrix as a C{scipy.sparse.csr_matrix}.
    """
    try:
        from scipy import sparse
    except ImportError:
        raise ImportError(
            "You should install scipy in order to use this function"
        ) from None

    edges = self.get_edgelist()
    if attribute is None:
        weights = [1] * len(edges)
    else:
        if attribute not in self.es.attribute_names():
            raise ValueError("Attribute does not exist")

        weights = self.es[attribute]

    N = self.vcount()
    mtx = sparse.csr_matrix((weights, list(zip(*edges))), shape=(N, N))

    if not self.is_directed():
        mtx = mtx + sparse.triu(mtx, 1).T + sparse.tril(mtx, -1).T
    return mtx


def _get_adjlist(self, mode="out", loops="twice", multiple=True):
    """Returns the adjacency list representation of the graph.

    The adjacency list representation is a list of lists. Each item of the
    outer list belongs to a single vertex of the graph. The inner list
    contains the neighbors of the given vertex.

    @param mode: if C{"out"}, returns the successors of the vertex. If
      C{"in"}, returns the predecessors of the vertex. If C{"all"}, both
      the predecessors and the successors will be returned. Ignored
      for undirected graphs.
    @param loops: whether to return loops in I{undirected} graphs once
      (C{"once"}), twice (C{"twice"}) or not at all (C{"ignore"}). C{False}
      is accepted as an alias to C{"ignore"} and C{True} is accepted as an
      alias to C{"twice"}. For directed graphs, C{"twice"} is equivalent
      to C{"once"} (except when C{mode} is C{"all"} because the graph is
      then treated as undirected).
    @param multiple: whether to return endpoints of multiple edges as many
      times as their multiplicities.
    """
    return [self.neighbors(idx, mode, loops, multiple) for idx in range(self.vcount())]


def _get_biadjacency(graph, types="type", *args, **kwds):
    """Returns the bipartite adjacency matrix of a bipartite graph. The
    bipartite adjacency matrix is an M{n} times M{m} matrix, where M{n} and
    M{m} are the number of vertices in the two vertex classes.

    @param types: a vector containing the vertex types, or an
      attribute name. Anything that evalulates to C{False} corresponds to
      vertices of the first kind, everything else to the second kind.
    @return: the bipartite adjacency matrix and two lists in a triplet. The
      first list defines the mapping between row indices of the matrix and the
      original vertex IDs. The second list is the same for the column indices.
    """
    # Deferred import to avoid cycles
    from igraph import Graph

    return super(Graph, graph).get_biadjacency(types, *args, **kwds)


def _get_inclist(graph, mode="out", loops="twice"):
    """Returns the incidence list representation of the graph.

    The incidence list representation is a list of lists. Each
    item of the outer list belongs to a single vertex of the graph.
    The inner list contains the IDs of the incident edges of the
    given vertex.

    @param mode: if C{"out"}, returns the successors of each vertex. If
      C{"in"}, returns the predecessors of each vertex. If C{"all"}, both
      the predecessors and the successors will be returned. Ignored
      for undirected graphs.
    @param loops: whether to return loops in I{undirected} graphs once
      (C{"once"}), twice (C{"twice"}) or not at all (C{"ignore"}). C{False}
      is accepted as an alias to C{"ignore"} and C{True} is accepted as an
      alias to C{"twice"}. For directed graphs, C{"twice"} is equivalent
      to C{"once"} (except when C{mode} is C{"all"} because the graph is
      then treated as undirected).
    """
    return [graph.incident(idx, mode, loops) for idx in range(graph.vcount())]



from igraph._igraph import (
    GraphBase,
    Vertex,
    Edge,
)
from igraph.seq import VertexSeq, EdgeSeq
from igraph.operators.functions import (
    disjoint_union,
    union,
    intersection,
)


__all__ = (
    "__iadd__",
    "__add__",
    "__and__",
    "__isub__",
    "__sub__",
    "__mul__",
    "__or__",
    "_disjoint_union",
    "_union",
    "_intersection",
)


def _disjoint_union(graph, other):
    """Creates the disjoint union of two (or more) graphs.

    @param other: graph or list of graphs to be united with the current one.
    @return: the disjoint union graph
    """
    if isinstance(other, GraphBase):
        other = [other]
    return disjoint_union([graph] + other)


def _union(graph, other, byname="auto"):
    """Creates the union of two (or more) graphs.

    @param other: graph or list of graphs to be united with the current one.
    @param byname: whether to use vertex names instead of ids. See
      L{igraph.operators.union} for details.
    @return: the union graph
    """
    if isinstance(other, GraphBase):
        other = [other]
    return union([graph] + other, byname=byname)


def _intersection(graph, other, byname="auto"):
    """Creates the intersection of two (or more) graphs.

    @param other: graph or list of graphs to be intersected with
      the current one.
    @param byname: whether to use vertex names instead of ids. See
      L{igraph.operators.intersection} for details.
    @return: the intersection graph
    """
    if isinstance(other, GraphBase):
        other = [other]
    return intersection([graph] + other, byname=byname)


def __iadd__(graph, other):
    """In-place addition (disjoint union).

    @see: L{__add__}
    """
    if isinstance(other, (int, str)):
        graph.add_vertices(other)
        return graph
    elif isinstance(other, tuple) and len(other) == 2:
        graph.add_edges([other])
        return graph
    elif isinstance(other, list):
        if not other:
            return graph
        if isinstance(other[0], tuple):
            graph.add_edges(other)
            return graph
        if isinstance(other[0], str):
            graph.add_vertices(other)
            return graph
    return NotImplemented


def __add__(graph, other):
    """Copies the graph and extends the copy depending on the type of
    the other object given.

    @param other: if it is an integer, the copy is extended by the given
      number of vertices. If it is a string, the copy is extended by a
      single vertex whose C{name} attribute will be equal to the given
      string. If it is a tuple with two elements, the copy
      is extended by a single edge. If it is a list of tuples, the copy
      is extended by multiple edges. If it is a L{Graph}, a disjoint
      union is performed.
    """
    # Deferred import to avoid cycles
    from igraph import Graph

    if isinstance(other, (int, str)):
        g = graph.copy()
        g.add_vertices(other)
    elif isinstance(other, tuple) and len(other) == 2:
        g = graph.copy()
        g.add_edges([other])
    elif isinstance(other, list):
        if len(other) > 0:
            if isinstance(other[0], tuple):
                g = graph.copy()
                g.add_edges(other)
            elif isinstance(other[0], str):
                g = graph.copy()
                g.add_vertices(other)
            elif isinstance(other[0], Graph):
                return graph.disjoint_union(other)
            else:
                return NotImplemented
        else:
            return graph.copy()

    elif isinstance(other, Graph):
        return graph.disjoint_union(other)
    else:
        return NotImplemented

    return g


def __and__(graph, other):
    """Graph intersection operator.

    @param other: the other graph to take the intersection with.
    @return: the intersected graph.
    """
    # Deferred import to avoid cycles
    from igraph import Graph

    if isinstance(other, Graph):
        return graph.intersection(other)
    else:
        return NotImplemented


def __isub__(graph, other):
    """In-place subtraction (difference).

    @see: L{__sub__}"""
    if isinstance(other, int):
        graph.delete_vertices([other])
    elif isinstance(other, tuple) and len(other) == 2:
        graph.delete_edges([other])
    elif isinstance(other, list):
        if len(other) > 0:
            if isinstance(other[0], tuple):
                graph.delete_edges(other)
            elif isinstance(other[0], (int, str)):
                graph.delete_vertices(other)
            else:
                return NotImplemented
    elif isinstance(other, Vertex):
        graph.delete_vertices(other)
    elif isinstance(other, VertexSeq):
        graph.delete_vertices(other)
    elif isinstance(other, Edge):
        graph.delete_edges(other)
    elif isinstance(other, EdgeSeq):
        graph.delete_edges(other)
    else:
        return NotImplemented
    return graph


def __sub__(graph, other):
    """Removes the given object(s) from the graph

    @param other: if it is an integer, removes the vertex with the given
      ID from the graph (note that the remaining vertices will get
      re-indexed!). If it is a tuple, removes the given edge. If it is
      a graph, takes the difference of the two graphs. Accepts
      lists of integers or lists of tuples as well, but they can't be
      mixed! Also accepts L{Edge} and L{EdgeSeq} objects.
    """
    # Deferred import to avoid cycles
    from igraph import Graph

    if isinstance(other, Graph):
        return graph.difference(other)

    result = graph.copy()
    if isinstance(other, (int, str)):
        result.delete_vertices([other])
    elif isinstance(other, tuple) and len(other) == 2:
        result.delete_edges([other])
    elif isinstance(other, list):
        if len(other) > 0:
            if isinstance(other[0], tuple):
                result.delete_edges(other)
            elif isinstance(other[0], (int, str)):
                result.delete_vertices(other)
            else:
                return NotImplemented
        else:
            return result
    elif isinstance(other, Vertex):
        result.delete_vertices(other)
    elif isinstance(other, VertexSeq):
        result.delete_vertices(other)
    elif isinstance(other, Edge):
        result.delete_edges(other)
    elif isinstance(other, EdgeSeq):
        result.delete_edges(other)
    else:
        return NotImplemented

    return result


def __mul__(graph, other):
    """Copies exact replicas of the original graph an arbitrary number of
    times.

    @param other: if it is an integer, multiplies the graph by creating the
      given number of identical copies and taking the disjoint union of
      them.
    """
    # Deferred import to avoid cycles
    from igraph import Graph

    if isinstance(other, int):
        if other == 0:
            return Graph()
        elif other == 1:
            return graph
        elif other > 1:
            return graph.disjoint_union([graph] * (other - 1))
        else:
            return NotImplemented

    return NotImplemented


def __or__(graph, other):
    """Graph union operator.

    @param other: the other graph to take the union with.
    @return: the union graph.
    """
    # Deferred import to avoid cycles
    from igraph import Graph

    if isinstance(other, Graph):
        return graph.union(other)
    else:
        return NotImplemented



"""
Drawing routines to draw graphs.

This module contains routines to draw graphs on:

  - Cairo surfaces (L{DefaultGraphDrawer})
  - Matplotlib axes (L{MatplotlibGraphDrawer})

It also contains routines to send an igraph graph directly to
(U{Cytoscape<http://www.cytoscape.org>}) using the
(U{CytoscapeRPC plugin<http://gforge.nbic.nl/projects/cytoscaperpc/>}), see
L{CytoscapeGraphDrawer}. L{CytoscapeGraphDrawer} can also fetch the current
network from Cytoscape and convert it to igraph format.
"""

from warnings import warn
from functools import wraps, partial

from igraph._igraph import convex_hull, VertexSeq
from igraph.drawing.baseclasses import AbstractGraphDrawer
from igraph.drawing.utils import FakeModule

from .edge import MatplotlibEdgeDrawer, EdgeCollection
from .polygon import HullCollection
from .utils import find_matplotlib
from .vertex import MatplotlibVertexDrawer, VertexCollection

__all__ = ("MatplotlibGraphDrawer",)

mpl, plt = find_matplotlib()
try:
    Artist = mpl.artist.Artist
    Affine2D = mpl.transforms.Affine2D
except AttributeError:
    Artist = FakeModule
    Affine2D = FakeModule

#####################################################################


# NOTE: https://github.com/networkx/grave/blob/main/grave/grave.py
def _stale_wrapper(func):
    """Decorator to manage artist state."""

    @wraps(func)
    def inner(self, *args, **kwargs):
        try:
            func(self, *args, **kwargs)
        finally:
            self.stale = False

    return inner


def _forwarder(forwards, cls=None):
    """Decorator to forward specific methods to Artist children."""
    if cls is None:
        return partial(_forwarder, forwards)

    def make_forward(name):
        def method(self, *args, **kwargs):
            ret = getattr(cls.mro()[1], name)(self, *args, **kwargs)
            for c in self.get_children():
                getattr(c, name)(*args, **kwargs)
            return ret

        return method

    for f in forwards:
        method = make_forward(f)
        method.__name__ = f
        method.__doc__ = "broadcasts {} to children".format(f)
        setattr(cls, f, method)

    return cls


def _additional_set_methods(attributes, cls=None):
    """Decorator to add specific set methods for children properties."""
    if cls is None:
        return partial(_additional_set_methods, attributes)

    def make_setter(name):
        def method(self, value):
            self.set(**{name: value})

        return method

    for attr in attributes:
        desc = attr.replace("_", " ")
        method = make_setter(attr)
        method.__name__ = f"set_{attr}"
        method.__doc__ = f"Set {desc}."
        setattr(cls, f"set_{attr}", method)

    return cls


@_additional_set_methods(
    (
        "vertex_color",
        "vertex_size",
        "vertex_font",
        "vertex_label",
        "vertex_label_angle",
        "vertex_label_color",
        "vertex_label_dist",
        "vertex_label_size",
        "vertex_order",
        "vertex_shape",
        "vertex_size",
        "edge_color",
        "edge_curved",
        "edge_font",
        "edge_arrow_size",
        "edge_arrow_width",
        "edge_width",
        "edge_label",
        "edge_background",
        "edge_align_label",
        "autocurve",
        "layout",
    )
)
@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_transform",
        "set_snap",
        "set_sketch_params",
        "set_figure",
        "set_animated",
        "set_picker",
    )
)
class GraphArtist(Artist, AbstractGraphDrawer):
    """Artist for an igraph.Graph object.

    @param graph: An igraph.Graph object to plot
    @param layout: A layout object or matrix of coordinates to use for plotting.
        Each element or row should describes the coordinates for a vertex.
    @param vertex_style: A dictionary specifying style options for vertices.
    @param edge_style: A dictionary specifying style options for edges.
    """

    def __init__(
        self,
        graph,
        vertex_drawer_factory=MatplotlibVertexDrawer,
        edge_drawer_factory=MatplotlibEdgeDrawer,
        mark_groups=None,
        layout=None,
        palette=None,
        **kwds,
    ):
        super().__init__()
        self.graph = graph
        self._vertex_drawer_factory = vertex_drawer_factory
        self._edge_drawer_factory = edge_drawer_factory
        self.kwds = kwds
        self.kwds["mark_groups"] = mark_groups
        self.kwds["palette"] = palette
        self.kwds["layout"] = layout

        self._kwds_post_update()

    def _kwds_post_update(self):
        self.kwds["layout"] = self.ensure_layout(self.kwds["layout"], self.graph)
        self._set_edge_curve()
        self._clear_state()
        self.stale = True

    def _clear_state(self):
        self._vertices = None
        self._edges = None
        self._vertex_labels = []
        self._edge_labels = []
        self._groups = None
        self._legend_info = {}

    def get_children(self):
        artists = []
        if self._groups is not None:
            artists.append(self._groups)
        # This way vertices are on top of edges, since they are drawn later
        if self._edges is not None:
            artists.append(self._edges)
        if self._vertices is not None:
            artists.append(self._vertices)
        artists.extend(self._edge_labels)
        artists.extend(self._vertex_labels)
        return tuple(artists)

    def _set_edge_curve(self):
        graph = self.graph
        kwds = self.kwds

        # Decide whether we need to calculate the curvature of edges
        # automatically -- and calculate them if needed.
        autocurve = kwds.get("autocurve", None)
        if autocurve or (
            autocurve is None
            and "edge_curved" not in kwds
            and "curved" not in self.graph.edge_attributes()
            and self.graph.ecount() < 10000
        ):
            from igraph import autocurve

            default = kwds.get("edge_curved", 0)
            if default is True:
                default = 0.5
            default = float(default)
            self.kwds["edge_curved"] = autocurve(
                graph,
                attribute=None,
                default=default,
            )

    def get_vertices(self):
        """Get VertexCollection artist."""
        return self._vertices

    def get_edges(self):
        """Get EdgeCollection artist."""
        return self._edges

    def get_groups(self):
        """Get HullCollection group/cluster/cover artists."""
        return self._groups

    def get_vertex_labels(self):
        """Get list of vertex label artists."""
        return self._vertex_labels

    def get_edge_labels(self):
        """Get list of edge label artists."""
        return self._edge_labels

    def get_datalim(self):
        """Get limits on x/y axes based on the graph layout data.

        There is a small padding based on the size of the vertex marker to
        ensure it fits into the canvas.
        """
        import numpy as np

        layout = self.kwds["layout"]

        if len(layout) == 0:
            mins = np.array([0, 0])
            maxs = np.array([1, 1])
            return (mins, maxs)

        # Use the layout as a base, and expand using bboxes from other artists
        mins = np.min(layout, axis=0).astype(float)
        maxs = np.max(layout, axis=0).astype(float)

        # NOTE: unlike other Collections, the vertices are basically a
        # PatchCollection with an offset transform using transData. Therefore,
        # care should be taken if one wants to include it here
        if self._vertices is not None:
            trans = self.axes.transData.transform
            trans_inv = self.axes.transData.inverted().transform
            verts = self._vertices
            for path, offset in zip(verts.get_paths(), verts._offsets):
                bbox = path.get_extents()
                mins = np.minimum(mins, trans_inv(bbox.min + trans(offset)))
                maxs = np.maximum(maxs, trans_inv(bbox.max + trans(offset)))

        if self._edges is not None:
            for path in self._edges.get_paths():
                bbox = path.get_extents()
                mins = np.minimum(mins, bbox.min)
                maxs = np.maximum(maxs, bbox.max)

        if self._groups is not None:
            for path in self._groups.get_paths():
                bbox = path.get_extents()
                mins = np.minimum(mins, bbox.min)
                maxs = np.maximum(maxs, bbox.max)

        # 5% padding, on each side
        pad = (maxs - mins) * 0.05
        mins -= pad
        maxs += pad

        return (mins, maxs)

    def _draw_vertex_labels(self):
        import numpy as np

        kwds = self.kwds
        layout = self.kwds["layout"]
        vertex_builder = self._vertex_builder
        vertex_order = self._vertex_order

        self._vertex_labels = []

        # Construct the iterator that we will use to draw the vertex labels
        if vertex_order is None:
            # Default vertex order
            vertex_coord_iter = zip(vertex_builder, layout)
        else:
            # Specified vertex order
            vertex_coord_iter = ((vertex_builder[i], layout[i]) for i in vertex_order)

        # Draw the vertex labels
        for vertex, coords in vertex_coord_iter:
            if vertex.label is None:
                continue

            label_size = kwds.get(
                "vertex_label_size",
                vertex.label_size,
            )

            label_color = kwds.get(
                "vertex_label_color",
                vertex.label_color,
            )

            # Locate text relative to vertex in data units. This is consistent
            # with the vertex size being in data units, but might be not fully
            # satisfactory when zooming in/out. In that case, revisit this
            # section
            dist = vertex.label_dist
            angle = vertex.label_angle
            if vertex.size is not None:
                vertex_width = vertex.size
                vertex_height = vertex.size
            else:
                vertex_width = vertex.width
                vertex_height = vertex.height
            xtext = dist * 0.5 * vertex_width * np.cos(angle)
            ytext = dist * 0.5 * vertex_height * np.sin(angle)
            xytext = (xtext, ytext)

            art = mpl.text.Annotation(
                vertex.label,
                coords,
                xytext=xytext,
                textcoords="offset points",
                fontsize=label_size,
                color=label_color,
                ha="center",
                va="center",
                clip_on=True,
                zorder=3,
            )
            self._vertex_labels.append(art)

    def _draw_edge_labels(self):
        graph = self.graph
        kwds = self.kwds
        vertex_builder = self._vertex_builder
        edge_builder = self._edge_builder
        edge_drawer = self._edge_drawer
        edge_order = self._edge_order or range(self.graph.ecount())

        self._edge_labels = []

        labels = kwds.get("edge_label", None)
        if labels is None:
            return

        edge_label_iter = (
            (labels[i], edge_builder[i], graph.es[i]) for i in edge_order
        )
        for label, visual_edge, edge in edge_label_iter:
            # Ask the edge drawer to propose an anchor point for the label
            src, dest = edge.tuple
            src_vertex, dest_vertex = vertex_builder[src], vertex_builder[dest]
            (x, y), (halign, valign) = edge_drawer.get_label_position(
                visual_edge,
                src_vertex,
                dest_vertex,
            )

            text_kwds = {}
            text_kwds["ha"] = halign.value
            text_kwds["va"] = valign.value

            if visual_edge.background is not None:
                text_kwds["bbox"] = {
                    "facecolor": visual_edge.background,
                    "edgecolor": "none",
                }
                text_kwds["ha"] = "center"
                text_kwds["va"] = "center"

            if visual_edge.align_label:
                # Rotate the text to align with the edge
                rotation = edge_drawer.get_label_rotation(
                    visual_edge,
                    src_vertex,
                    dest_vertex,
                )
                text_kwds["rotation"] = rotation

            art = mpl.text.Annotation(
                label,
                (x, y),
                fontsize=visual_edge.label_size,
                color=visual_edge.label_color,
                transform=self.axes.transData,
                clip_on=True,
                zorder=3,
                **text_kwds,
            )
            self._edge_labels.append(art)

    def _draw_groups(self):
        """Draw the highlighted vertex groups, if requested"""
        # Deferred import to avoid a cycle in the import graph
        from igraph.clustering import VertexClustering, VertexCover

        mark_groups = self.kwds["mark_groups"]
        if not mark_groups:
            return

        kwds = self.kwds
        palette = self.kwds["palette"]
        layout = self.kwds["layout"]
        vertex_builder = self._vertex_builder

        # Figure out what to do with mark_groups in order to be able to
        # iterate over it and get memberlist-color pairs
        if isinstance(mark_groups, dict):
            # Dictionary mapping vertex indices or tuples of vertex
            # indices to colors
            group_iter = iter(mark_groups.items())
        elif isinstance(mark_groups, (VertexClustering, VertexCover)):
            # Vertex clustering
            group_iter = ((group, color) for color, group in enumerate(mark_groups))
        elif hasattr(mark_groups, "__iter__"):
            # One-off generators: we need to store the actual list for future
            # calls (e.g. resizing, recoloring, etc.). If we don't do this,
            # the generator is exhausted: we cannot rewind it.
            self.mark_groups = mark_groups = list(mark_groups)
            # Lists, tuples, iterators etc
            group_iter = iter(mark_groups)
        else:
            # False
            group_iter = iter({}.items())

        if kwds.get("legend", False):
            legend_info = {
                "handles": [],
                "labels": [],
            }

        # Iterate over color-memberlist pairs
        polygons = []
        corner_radii = []
        facecolors = []
        edgecolors = []
        for group, color_id in group_iter:
            if not group or color_id is None:
                continue

            color = palette.get(color_id)

            if isinstance(group, VertexSeq):
                group = [vertex.index for vertex in group]
            if not hasattr(group, "__iter__"):
                raise TypeError("group membership list must be iterable")

            # Get the vertex indices that constitute the convex hull
            hull = [group[i] for i in convex_hull([layout[idx] for idx in group])]

            # Construct the hull polygon
            polygon = [layout[idx] for idx in hull]

            # Calculate rounding radius and facecolor
            corner_radius = 1.25 * max(vertex_builder[idx].size for idx in hull)
            facecolor = (color[0], color[1], color[2], 0.25 * color[3])

            if kwds.get("legend", False):
                legend_info["handles"].append(
                    plt.Rectangle(
                        (0, 0),
                        0,
                        0,
                        facecolor=facecolor,
                        edgecolor=color,
                    )
                )
                legend_info["labels"].append(str(color_id))

            if len(polygon) >= 1:
                polygons.append(mpl.path.Path(polygon))
                corner_radii.append(corner_radius)
                facecolors.append(facecolor)
                edgecolors.append(color)

        art = HullCollection(
            polygons,
            corner_radius=corner_radii,
            facecolor=facecolors,
            edgecolor=edgecolors,
            transform=self.axes.transData,
        )
        self._groups = art

        if kwds.get("legend", False):
            self.legend_info = legend_info

    def _draw_vertices(self):
        """Draw the vertices"""
        graph = self.graph
        layout = self.kwds["layout"]
        vertex_drawer = self._vertex_drawer
        vertex_builder = self._vertex_builder
        vertex_order = self._vertex_order

        vs = graph.vs
        if vertex_order is None:
            # Default vertex order
            vertex_coord_iter = zip(vs, vertex_builder, layout)
        else:
            # Specified vertex order
            vertex_coord_iter = (
                (vs[i], vertex_builder[i], layout[i]) for i in vertex_order
            )
        offsets = []
        patches = []
        for vertex, visual_vertex, coords in vertex_coord_iter:
            art = vertex_drawer.draw(visual_vertex, vertex, coords)
            patches.append(art)
            offsets.append(list(coords))

        art = VertexCollection(
            patches,
            offsets=offsets if offsets else None,
            offset_transform=self.axes.transData,
            match_original=True,
            transform=Affine2D(),
        )
        self._vertices = art

    def _draw_edges(self):
        """Draw the edges"""
        graph = self.graph
        vertex_builder = self._vertex_builder
        edge_drawer = self._edge_drawer
        edge_builder = self._edge_builder
        edge_order = self._edge_order

        es = graph.es
        if edge_order is None:
            # Default edge order
            edge_coord_iter = zip(es, edge_builder)
        else:
            # Specified edge order
            edge_coord_iter = ((es[i], edge_builder[i]) for i in edge_order)

        directed = graph.is_directed()

        visual_vertices = []
        edgepatches = []
        arrow_sizes = []
        arrow_widths = []
        loop_sizes = []
        curved = []
        for edge, visual_edge in edge_coord_iter:
            edge_vertices = [vertex_builder[v] for v in edge.tuple]
            art = edge_drawer.build_patch(visual_edge, *edge_vertices)
            edgepatches.append(art)
            visual_vertices.append(edge_vertices)
            arrow_sizes.append(visual_edge.arrow_size)
            arrow_widths.append(visual_edge.arrow_width)
            loop_sizes.append(visual_edge.loop_size)
            curved.append(visual_edge.curved)

        art = EdgeCollection(
            edgepatches,
            visual_vertices=visual_vertices,
            directed=directed,
            arrow_sizes=arrow_sizes,
            arrow_widths=arrow_widths,
            loop_sizes=loop_sizes,
            curved=curved,
            transform=self.axes.transData,
        )
        self._edges = art

    def _reprocess(self):
        """Prepare artist and children for the actual drawing.

        Children are not drawn here, but the dictionaries of properties are
        marshalled to their specific artists.
        """
        # clear state and mark as stale
        # since all children artists are part of the state, clearing it
        # will trigger a deletion by the backend at the next draw cycle
        self._clear_state()
        self.stale = True

        # get local refs to everything (just for less typing)
        graph = self.graph
        palette = self.kwds["palette"]
        layout = self.kwds["layout"]
        kwds = self.kwds

        # Construct the vertex, edge and label drawers
        if not hasattr(self, "_vertex_drawer"):
            self._vertex_drawer = self._vertex_drawer_factory(
                self.axes, palette, layout
            )
        if not hasattr(self, "_edge_drawer"):
            self._edge_drawer = self._edge_drawer_factory(self.axes, palette)

        # Construct the visual vertex/edge builders based on the specifications
        # provided by the vertex_drawer and the edge_drawer
        if not hasattr(self, "_vertex_builder"):
            self._vertex_builder = self._vertex_drawer.VisualVertexBuilder(
                graph.vs, kwds
            )
        if not hasattr(self, "_edge_builder"):
            self._edge_builder = self._edge_drawer.VisualEdgeBuilder(graph.es, kwds)

        # Determine the order in which we will draw the vertices and edges
        # These methods come from AbstractGraphDrawer
        self._vertex_order = self._determine_vertex_order(graph, kwds)
        self._edge_order = self._determine_edge_order(graph, kwds)

        self._draw_groups()
        self._draw_vertices()
        self._draw_edges()
        self._draw_vertex_labels()
        self._draw_edge_labels()

        # Callbacks for other vertex properties, to ensure they are in sync
        # with vertex_builder.
        # NOTE: no need to reprocess here because it does not affect other
        # parts of the container artist (e.g. edges)
        def vertex_stale_callback(artist):
            # If the stale state emerges from other properties, we can salvage
            # the other artists but we have to update the vertex builder anyway
            # in case a _reprocess is triggered by something else.
            prop_pairs = (
                ("edgecolor", "frame_color"),
                ("facecolor", "color"),
                ("linewidth", "frame_width"),
                ("zorder", "zorder"),
                ("sizes", "size"),
            )
            for mpl_prop, ig_prop in prop_pairs:
                values = getattr(artist, "get_" + mpl_prop)()
                try:
                    iter(values)
                except TypeError:
                    values = [values] * len(artist.get_paths())
                for value, visual_vertex in zip(values, self._vertex_builder):
                    setattr(visual_vertex, ig_prop, value)

            # If the size is stale, one needs to redraw everything
            if artist._stale_size:
                self._reprocess()

        # Edge callback, keeps the edge builder in sync with the actual state
        # of the artist
        def edge_stale_callback(artist):
            prop_pairs = (
                ("edgecolor", "color"),
                ("linewidth", "width"),
                ("zorder", "zorder"),
                ("arrow_size", "arrow_size"),
                ("arrow_width", "arrow_width"),
            )
            for mpl_prop, ig_prop in prop_pairs:
                values = getattr(artist, "get_" + mpl_prop)()
                try:
                    iter(values)
                except TypeError:
                    values = [values] * len(artist.get_paths())
                for value, visual_edge in zip(values, self._edge_builder):
                    setattr(visual_edge, ig_prop, value)

                # Sync facecolor from edgecolor
                if mpl_prop == "edgecolor":
                    artist._facecolors = artist._edgecolors

        self._vertices.stale_callback_post = vertex_stale_callback
        self._edges.stale_callback_post = edge_stale_callback

        # Forward mpl properties to children
        # TODO sort out all of the things that need to be forwarded
        for child in self.get_children():
            # set the figure / axes on child, this ensures each primitive
            # knows where to draw
            if hasattr(child, "set_figure"):
                child.set_figure(self.figure)
            child.axes = self.axes

            # forward the clippath/box to the children need this logic
            # because mpl exposes some fast-path logic
            clip_path = self.get_clip_path()
            if clip_path is None:
                clip_box = self.get_clip_box()
                child.set_clip_box(clip_box)
            else:
                child.set_clip_path(clip_path)

    @_stale_wrapper
    def draw(self, renderer, *args, **kwds):
        """Draw each of the children, with some buffering mechanism."""
        if not self.get_visible():
            return

        if not self.get_children():
            self._reprocess()

        # NOTE: looks like we have to manage the zorder ourselves
        children = list(self.get_children())
        children.sort(key=lambda x: x.zorder)
        for art in children:
            art.draw(renderer, *args, **kwds)

    def set(
        self,
        **kwds,
    ):
        """Set multiple parameters at once.

        The same options can be used as in the igraph.plot function.
        """
        if len(kwds) == 0:
            return

        self.kwds.update(kwds)
        self._kwds_post_update()

    def contains(self, mouseevent):
        """Track 'contains' event for mouse interactions."""
        props = {"vertices": [], "edges": []}
        hit = False
        for i, art in enumerate(self._edges):
            edge_hit = art.contains(mouseevent)[0]
            hit |= edge_hit
            props["edges"].append(i)

        for i, art in enumerate(self._vertices):
            vertex_hit = art.contains(mouseevent)[0]
            hit |= vertex_hit
            props["vertices"].append(i)

        return hit, props

    def pick(self, mouseevent):
        """Track 'pick' event for mouse interactions."""
        if self.pickable():
            picker = self.get_picker()
            if callable(picker):
                inside, prop = picker(self, mouseevent)
            else:
                inside, prop = self.contains(mouseevent)
            if inside:
                self.figure.canvas.pick_event(mouseevent, self, **prop)


class MatplotlibGraphDrawer(AbstractGraphDrawer):
    """Graph drawer that uses a pyplot.Axes as context"""

    _shape_dict = {
        "rectangle": "s",
        "circle": "o",
        "hidden": "none",
        "triangle-up": "^",
        "triangle-down": "v",
    }

    def __init__(
        self,
        ax,
        vertex_drawer_factory=MatplotlibVertexDrawer,
        edge_drawer_factory=MatplotlibEdgeDrawer,
    ):
        """Constructs the graph drawer and associates it with the mpl Axes

        @param ax: the matplotlib Axes to draw into.
        @param vertex_drawer_factory: a factory method that returns an
            L{AbstractVertexDrawer} instance bound to the given Matplotlib axes.
            The factory method must take three parameters: the axes and the
            palette to be used for drawing colored vertices, and the layout of
            the graph. The default vertex drawer is L{MatplotlibVertexDrawer}.
        @param edge_drawer_factory: a factory method that returns an
            L{AbstractEdgeDrawer} instance bound to a given Matplotlib Axes.
            The factory method must take two parameters: the Axes and the palette
            to be used for drawing colored edges. The default edge drawer is
            L{MatplotlibEdgeDrawer}.
        """
        self.ax = ax
        self.vertex_drawer_factory = vertex_drawer_factory
        self.edge_drawer_factory = edge_drawer_factory

    def draw(self, graph, *args, **kwds):
        if args:
            warn(
                "Positional arguments to plot functions are ignored "
                "and will be deprecated soon.",
                DeprecationWarning,
                stacklevel=1,
            )

        # Some abbreviations for sake of simplicity
        ax = self.ax

        # Create artist
        art = GraphArtist(
            graph,
            vertex_drawer_factory=self.vertex_drawer_factory,
            edge_drawer_factory=self.edge_drawer_factory,
            *args,  # noqa: B026
            **kwds,
        )

        # Bind artist to axes
        ax.add_artist(art)

        # Create children artists (this also binds them to the axes)
        art._reprocess()

        # Legend for groups
        if ("mark_groups" in kwds) and kwds.get("legend", False):
            ax.legend(
                art._legend_info["handles"],
                art._legend_info["labels"],
            )

        # Set new data limits
        ax.update_datalim(art.get_datalim())

        # Despine
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.spines["bottom"].set_visible(False)

        # Remove axis ticks
        ax.set_xticks([])
        ax.set_yticks([])

        # Autoscale for x/y axis limits
        ax.autoscale_view()

        return art



import unittest

from igraph import Graph, disjoint_union, intersection, union

try:
    import numpy as np
except ImportError:
    np = None


class OperatorTests(unittest.TestCase):
    def testComplementer(self):
        g = Graph.Full(3)
        g2 = g.complementer()
        self.assertTrue(g2.vcount() == 3 and g2.ecount() == 3)
        self.assertTrue(sorted(g2.get_edgelist()) == [(0, 0), (1, 1), (2, 2)])

        g = Graph.Full(3) + Graph.Full(2)
        g2 = g.complementer(False)
        self.assertTrue(
            sorted(g2.get_edgelist())
            == [(0, 3), (0, 4), (1, 3), (1, 4), (2, 3), (2, 4)]
        )

        g2 = g.complementer(loops=True)
        self.assertTrue(
            sorted(g2.get_edgelist())
            == [
                (0, 0),
                (0, 3),
                (0, 4),
                (1, 1),
                (1, 3),
                (1, 4),
                (2, 2),
                (2, 3),
                (2, 4),
                (3, 3),
                (4, 4),
            ]
        )

    def testMultiplication(self):
        g = Graph.Full(3) * 3
        self.assertTrue(
            g.vcount() == 9
            and g.ecount() == 9
            and g.connected_components().membership == [0, 0, 0, 1, 1, 1, 2, 2, 2]
        )

    def testDifference(self):
        g = Graph.Tree(7, 2) - Graph.Lattice([7])
        self.assertTrue(g.vcount() == 7 and g.ecount() == 5)
        self.assertTrue(
            sorted(g.get_edgelist()) == [(0, 2), (1, 3), (1, 4), (2, 5), (2, 6)]
        )

    def testDifferenceWithSelfLoop(self):
        # https://github.com/igraph/igraph/issues/597#
        g = Graph.Ring(10) + [(0, 0)]
        g -= Graph.Ring(5)
        self.assertTrue(g.vcount() == 10 and g.ecount() == 7)
        self.assertTrue(
            sorted(g.get_edgelist())
            == [(0, 0), (0, 9), (4, 5), (5, 6), (6, 7), (7, 8), (8, 9)]
        )

    def testDisjointUnion(self):
        g1 = Graph.Tree(7, 2)
        g2 = Graph.Lattice([7])

        # Method
        g = g1.disjoint_union(g2)
        self.assertTrue(g.vcount() == 14 and g.ecount() == 13)

        # Module function
        g = disjoint_union([g1, g2])
        self.assertTrue(g.vcount() == 14 and g.ecount() == 13)

    def testDisjointUnionNoGraphs(self):
        self.assertRaises(ValueError, disjoint_union, [])

    def testDisjointUnionSingle(self):
        g1 = Graph.Tree(7, 2)
        g = disjoint_union([g1])
        self.assertTrue(g != g1)
        self.assertTrue(g.vcount() == g1.vcount() and g.ecount() == g1.ecount())
        self.assertTrue(g.is_directed() == g1.is_directed())
        self.assertTrue(g.get_edgelist() == g1.get_edgelist())

    def testUnion(self):
        g = Graph.Tree(7, 2) | Graph.Lattice([7])
        self.assertTrue(g.vcount() == 7 and g.ecount() == 12)
        self.assertTrue(
            sorted(g.get_edgelist())
            == [
                (0, 1),
                (0, 2),
                (0, 6),
                (1, 2),
                (1, 3),
                (1, 4),
                (2, 3),
                (2, 5),
                (2, 6),
                (3, 4),
                (4, 5),
                (5, 6),
            ]
        )

    def testUnionWithConflict(self):
        g1 = Graph.Tree(7, 2)
        g1["name"] = "Tree"
        g2 = Graph.Lattice([7])
        g2["name"] = "Lattice"
        g = union([g1, g2])  # Issue 422
        self.assertTrue(
            sorted(g.get_edgelist())
            == [
                (0, 1),
                (0, 2),
                (0, 6),
                (1, 2),
                (1, 3),
                (1, 4),
                (2, 3),
                (2, 5),
                (2, 6),
                (3, 4),
                (4, 5),
                (5, 6),
            ]
        )
        self.assertTrue(
            sorted(g.attributes()),
            ["name_1", "name_2"],
        )

    def testUnionMethod(self):
        g = Graph.Tree(7, 2).union(Graph.Lattice([7]))
        self.assertTrue(g.vcount() == 7 and g.ecount() == 12)

    def testUnionNoGraphs(self):
        self.assertRaises(ValueError, union, [])

    def testUnionSingle(self):
        g1 = Graph.Tree(7, 2)
        g = union([g1])
        self.assertTrue(g != g1)
        self.assertTrue(g.vcount() == g1.vcount() and g.ecount() == g1.ecount())
        self.assertTrue(g.is_directed() == g1.is_directed())
        self.assertTrue(g.get_edgelist() == g1.get_edgelist())

    def testUnionMany(self):
        gs = [Graph.Tree(7, 2), Graph.Lattice([7]), Graph.Lattice([7])]
        g = union(gs)
        self.assertTrue(g.vcount() == 7 and g.ecount() == 12)

    def testUnionManyAttributes(self):
        gs = [
            Graph.Formula("A-B"),
            Graph.Formula("A-B,C-D"),
        ]
        gs[0]["attr"] = "graph1"
        gs[0].vs["attr"] = ["set", "set_too"]
        gs[0].vs["attr2"] = ["set", "set_too"]
        gs[1].vs[0]["attr"] = "set"
        gs[1].vs[0]["attr2"] = "conflict"
        g = union(gs)
        names = g.vs["name"]
        self.assertTrue(g["attr"] == "graph1")
        self.assertTrue(g.vs[names.index("A")]["attr"] == "set")
        self.assertTrue(g.vs[names.index("B")]["attr"] == "set_too")
        self.assertTrue(g.ecount() == 2)
        self.assertTrue(
            sorted(g.vertex_attributes()) == ["attr", "attr2_1", "attr2_2", "name"]
        )

    def testUnionManyEdgemap(self):
        gs = [
            Graph.Formula("A-B"),
            Graph.Formula("C-D, A-B"),
        ]
        gs[0].es[0]["attr"] = "set"
        gs[1].es[0]["attr"] = "set_too"
        g = union(gs)
        for e in g.es:
            vnames = [g.vs[e.source]["name"], g.vs[e.target]["name"]]
            if set(vnames) == {"A", "B"}:
                self.assertTrue(e["attr"] == "set")
            else:
                self.assertTrue(e["attr"] == "set_too")

    def testIntersection(self):
        g = Graph.Tree(7, 2) & Graph.Lattice([7])
        self.assertTrue(g.get_edgelist() == [(0, 1)])

    def testIntersectionMethod(self):
        g = Graph.Tree(7, 2).intersection(Graph.Lattice([7]))
        self.assertTrue(g.get_edgelist() == [(0, 1)])

    def testIntersectionNoGraphs(self):
        self.assertRaises(ValueError, intersection, [])

    def testIntersectionSingle(self):
        g1 = Graph.Tree(7, 2)
        g = intersection([g1])
        self.assertTrue(g != g1)
        self.assertTrue(g.vcount() == g1.vcount() and g.ecount() == g1.ecount())
        self.assertTrue(g.is_directed() == g1.is_directed())
        self.assertTrue(g.get_edgelist() == g1.get_edgelist())

    def testIntersectionMany(self):
        gs = [Graph.Tree(7, 2), Graph.Lattice([7])]
        g = intersection(gs)
        self.assertTrue(g.get_edgelist() == [(0, 1)])

    def testIntersectionManyAttributes(self):
        gs = [Graph.Tree(7, 2), Graph.Lattice([7])]
        gs[0]["attr"] = "graph1"
        gs[0].vs["name"] = ["one", "two", "three", "four", "five", "six", "7"]
        gs[1].vs["name"] = ["one", "two", "three", "four", "five", "six", "7"]
        gs[0].vs[0]["attr"] = "set"
        gs[1].vs[5]["attr"] = "set_too"
        g = intersection(gs)
        names = g.vs["name"]
        self.assertTrue(g["attr"] == "graph1")
        self.assertTrue(g.vs[names.index("one")]["attr"] == "set")
        self.assertTrue(g.vs[names.index("six")]["attr"] == "set_too")
        self.assertTrue(g.ecount() == 1)
        self.assertTrue(
            set(g.get_edgelist()[0]) == {names.index("one"), names.index("two")},
        )

    def testIntersectionManyEdgemap(self):
        gs = [
            Graph.Formula("A-B"),
            Graph.Formula("A-B,C-D"),
        ]
        gs[0].es[0]["attr"] = "set"
        gs[1].es[1]["attr"] = "set_too"
        g = intersection(gs)
        self.assertTrue(g.es["attr"] == ["set"])

    def testInPlaceAddition(self):
        g = Graph.Full(3)
        orig = g

        # Adding vertices
        g += 2
        self.assertTrue(
            g.vcount() == 5
            and g.ecount() == 3
            and g.connected_components().membership == [0, 0, 0, 1, 2]
        )

        # Adding a vertex by name
        g += "spam"
        self.assertTrue(
            g.vcount() == 6
            and g.ecount() == 3
            and g.connected_components().membership == [0, 0, 0, 1, 2, 3]
        )

        # Adding a single edge
        g += (2, 3)
        self.assertTrue(
            g.vcount() == 6
            and g.ecount() == 4
            and g.connected_components().membership == [0, 0, 0, 0, 1, 2]
        )

        # Adding two edges
        g += [(3, 4), (2, 4), (4, 5)]
        self.assertTrue(
            g.vcount() == 6
            and g.ecount() == 7
            and g.connected_components().membership == [0] * 6
        )

        # Adding two more vertices
        g += ["eggs", "bacon"]
        self.assertEqual(
            g.vs["name"], [None, None, None, None, None, "spam", "eggs", "bacon"]
        )

        # Did we really use the original graph so far?
        # TODO: disjoint union should be modified so that this assertion
        # could be moved to the end
        self.assertTrue(id(g) == id(orig))

        # Adding another graph
        g += Graph.Full(3)
        self.assertTrue(
            g.vcount() == 11
            and g.ecount() == 10
            and g.connected_components().membership == [0, 0, 0, 0, 0, 0, 1, 2, 3, 3, 3]
        )

        # Adding two graphs
        g += [Graph.Full(3), Graph.Full(2)]
        self.assertTrue(
            g.vcount() == 16
            and g.ecount() == 14
            and g.connected_components().membership
            == [0, 0, 0, 0, 0, 0, 1, 2, 3, 3, 3, 4, 4, 4, 5, 5]
        )

    def testAddition(self):
        g0 = Graph.Full(3)

        # Adding vertices
        g = g0 + 2
        self.assertTrue(
            g.vcount() == 5
            and g.ecount() == 3
            and g.connected_components().membership == [0, 0, 0, 1, 2]
        )
        g0 = g

        # Adding vertices by name
        g = g0 + "spam"
        self.assertTrue(
            g.vcount() == 6
            and g.ecount() == 3
            and g.connected_components().membership == [0, 0, 0, 1, 2, 3]
        )
        g0 = g

        # Adding a single edge
        g = g0 + (2, 3)
        self.assertTrue(
            g.vcount() == 6
            and g.ecount() == 4
            and g.connected_components().membership == [0, 0, 0, 0, 1, 2]
        )
        g0 = g

        # Adding two edges
        g = g0 + [(3, 4), (2, 4), (4, 5)]
        self.assertTrue(
            g.vcount() == 6
            and g.ecount() == 7
            and g.connected_components().membership == [0] * 6
        )
        g0 = g

        # Adding another graph
        g = g0 + Graph.Full(3)
        self.assertTrue(
            g.vcount() == 9
            and g.ecount() == 10
            and g.connected_components().membership == [0, 0, 0, 0, 0, 0, 1, 1, 1]
        )

    def testInPlaceSubtraction(self):
        g = Graph.Full(8)
        orig = g

        # Deleting a vertex by vertex selector
        g -= 7
        self.assertTrue(
            g.vcount() == 7
            and g.ecount() == 21
            and g.connected_components().membership == [0, 0, 0, 0, 0, 0, 0]
        )

        # Deleting a vertex
        g -= g.vs[6]
        self.assertTrue(
            g.vcount() == 6
            and g.ecount() == 15
            and g.connected_components().membership == [0, 0, 0, 0, 0, 0]
        )

        # Deleting two vertices
        g -= [4, 5]
        self.assertTrue(
            g.vcount() == 4
            and g.ecount() == 6
            and g.connected_components().membership == [0, 0, 0, 0]
        )

        # Deleting an edge
        g -= (1, 2)
        self.assertTrue(
            g.vcount() == 4
            and g.ecount() == 5
            and g.connected_components().membership == [0, 0, 0, 0]
        )

        # Deleting three more edges
        g -= [(1, 3), (0, 2), (0, 3)]
        self.assertTrue(
            g.vcount() == 4
            and g.ecount() == 2
            and g.connected_components().membership == [0, 0, 1, 1]
        )

        # Did we really use the original graph so far?
        self.assertTrue(id(g) == id(orig))

        # Subtracting a graph
        g2 = Graph.Tree(3, 2)
        g -= g2
        self.assertTrue(
            g.vcount() == 4
            and g.ecount() == 1
            and g.connected_components().membership == [0, 1, 2, 2]
        )

    def testNonzero(self):
        self.assertTrue(Graph(1))
        self.assertFalse(Graph(0))

    def testLength(self):
        self.assertRaises(TypeError, len, Graph(15))
        self.assertTrue(len(Graph(15).vs) == 15)
        self.assertTrue(len(Graph.Full(5).es) == 10)

    def testSimplify(self):
        el = [(0, 1), (1, 0), (1, 2), (2, 3), (2, 3), (2, 3), (3, 3)]
        g = Graph(el)
        g.es["weight"] = [1, 2, 3, 4, 5, 6, 7]

        g2 = g.copy()
        g2.simplify()
        self.assertTrue(g2.vcount() == g.vcount())
        self.assertTrue(g2.ecount() == 3)

        g2 = g.copy()
        g2.simplify(loops=False)
        self.assertTrue(g2.vcount() == g.vcount())
        self.assertTrue(g2.ecount() == 4)

        g2 = g.copy()
        g2.simplify(multiple=False)
        self.assertTrue(g2.vcount() == g.vcount())
        self.assertTrue(g2.ecount() == g.ecount() - 1)

    def testContractVertices(self):
        g = Graph.Full(4) + Graph.Full(4) + [(0, 5), (1, 4)]

        g2 = g.copy()
        g2.contract_vertices([0, 1, 2, 3, 1, 0, 4, 5])
        self.assertEqual(g2.vcount(), 6)
        self.assertEqual(g2.ecount(), g.ecount())
        self.assertEqual(
            sorted(g2.get_edgelist()),
            [
                (0, 0),
                (0, 1),
                (0, 1),
                (0, 2),
                (0, 3),
                (0, 4),
                (0, 5),
                (1, 1),
                (1, 2),
                (1, 3),
                (1, 4),
                (1, 5),
                (2, 3),
                (4, 5),
            ],
        )

        g2 = g.copy()
        g2.contract_vertices([0, 1, 2, 3, 1, 0, 6, 7])
        self.assertEqual(g2.vcount(), 8)
        self.assertEqual(g2.ecount(), g.ecount())
        self.assertEqual(
            sorted(g2.get_edgelist()),
            [
                (0, 0),
                (0, 1),
                (0, 1),
                (0, 2),
                (0, 3),
                (0, 6),
                (0, 7),
                (1, 1),
                (1, 2),
                (1, 3),
                (1, 6),
                (1, 7),
                (2, 3),
                (6, 7),
            ],
        )

        g2 = Graph(10)
        g2.contract_vertices([0, 0, 1, 1, 2, 2, 3, 3, 4, 4])
        self.assertEqual(g2.vcount(), 5)
        self.assertEqual(g2.ecount(), 0)

    @unittest.skipIf(np is None, "test case depends on NumPy")
    def testContractVerticesWithNumPyIntegers(self):
        g = Graph.Full(4) + Graph.Full(4) + [(0, 5), (1, 4)]
        g2 = g.copy()
        g2.contract_vertices([np.int32(x) for x in [0, 1, 2, 3, 1, 0, 6, 7]])
        self.assertEqual(g2.vcount(), 8)
        self.assertEqual(g2.ecount(), g.ecount())
        self.assertEqual(
            sorted(g2.get_edgelist()),
            [
                (0, 0),
                (0, 1),
                (0, 1),
                (0, 2),
                (0, 3),
                (0, 6),
                (0, 7),
                (1, 1),
                (1, 2),
                (1, 3),
                (1, 6),
                (1, 7),
                (2, 3),
                (6, 7),
            ],
        )

    def testReverseEdges(self):
        g = Graph.Tree(10, 3, mode="out")
        g.reverse_edges([0, 1, 2])
        self.assertEqual(
            g.get_edgelist(),
            [(1, 0), (2, 0), (3, 0), (1, 4), (1, 5), (1, 6), (2, 7), (2, 8), (2, 9)],
        )

        g = Graph.Tree(13, 3, mode="in")
        g.reverse_edges()
        self.assertTrue(g.isomorphic(Graph.Tree(13, 3, mode="out")))


def suite():
    operator_suite = unittest.defaultTestLoader.loadTestsFromTestCase(OperatorTests)
    return unittest.TestSuite([operator_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()
